'''
Created on Jun 8, 2021

@author: mballance
'''
import os
import sys


def find_project_dir():
    """Attempt to find the real project directory"""
    cwd = os.getcwd()
        
    if os.path.isdir(os.path.join(cwd, "packages")):
        return cwd
    else:
        # Go up the path
        parent = os.path.dirname(cwd)
        while parent != "" and parent != "/":
            if os.path.isdir(os.path.join(parent, "packages")):
                return parent
            parent = os.path.dirname(parent)
            
    return None

def ensure_have_project_dir(args):
    if args.project_dir is None:
        print("Note: Attempting to discover project_dir")
        sys.stdout.flush()
        args.project_dir = find_project_dir()
        
        if args.project_dir is None:
            raise Exception("Failed to find project_dir ; specify with --project-dir")
        else:
            print("Note: project_dir is " + args.project_dir) 
            sys.stdout.flush()



#!/bin/python
import os.path
import sys
import subprocess
import ivpm

# Subcommands
# update: 
# - discovers the project location by finding ivpm.info
# - locates the packages.info file
# - Processes the existing packages (if any)
#   - Load the packages.mf file from the packages directory
# - For each package in the package.info file:
#   - If the selected package is not source
#     - Locate the package.mf file from the repositories
#   - Otherwise,
#
#   - Check the package vers 

# Discover location
# Need to determine the project that we're in


#********************************************************************
# ivpm_main()
# This is the 
#********************************************************************
def ivpm_main(project_dir, argv):
    """Back-compat entry point for the bootstrap ivpm.py stored in projects"""
    
    # Determine where the ivpm package directory is
    scripts_dir = os.path.dirname(os.path.realpath(__file__))
    ivpm_dir = os.path.dirname(scripts_dir)
    
    sys.path.insert(0, os.path.join(ivpm_dir, "src"))
    from ivpm.__main__ import main
    main(project_dir)




'''
Created on Jan 19, 2020

@author: ballance
'''

import argparse
import os
import sys

from typing import Dict, List, Tuple

from ivpm.packages_info import PackagesInfo
from ivpm.proj_info import ProjInfo
from .cmds.cmd_activate import CmdActivate
from .cmds.cmd_build import CmdBuild
from .cmds.cmd_cache import CmdCache
from .cmds.cmd_init import CmdInit
from .cmds.cmd_update import CmdUpdate
from .cmds.cmd_clone import CmdClone
from .cmds.cmd_git_status import CmdGitStatus
from .cmds.cmd_git_update import CmdGitUpdate
from .cmds.cmd_pkg_info import CmdPkgInfo
from .cmds.cmd_share import CmdShare
from .cmds.cmd_snapshot import CmdSnapshot
from .cmds.cmd_status import CmdStatus
from .cmds.cmd_sync import CmdSync


def get_parser(parser_ext : List = None, options_ext : List = None):
    """Create the argument parser"""
    subcommands : Dict[str, object] = {}
    parser = argparse.ArgumentParser(prog="ivpm")
    
    subparser = parser.add_subparsers()
    subparser.required = True
    subparser.dest = 'command'

    activate_cmd = subparser.add_parser("activate",
        help="Starts a new shell that contains the activated python virtual environment")
    activate_cmd.add_argument("-c",
        help="When specified, executes the specified string")
    activate_cmd.add_argument("-p", "--project-dir", dest="project_dir",
        help="Specifies the project directory to use (default: cwd)")
    activate_cmd.add_argument("args", nargs='*')
    activate_cmd.set_defaults(func=CmdActivate())
    subcommands["activate"] = activate_cmd

    build_cmd = subparser.add_parser("build",
        help="Build all sub-projects with an IVPM-supported build infrastructure (Python)")
    build_cmd.add_argument("-d", "--dep-set", dest="dep_set", 
        help="Uses dependencies from specified dep-set instead of 'default-dev'")
    build_cmd.add_argument("-g", "--debug", 
        action="store_true",
        help="Enables debug for native extensions")
    build_cmd.set_defaults(func=CmdBuild())
    subcommands["build"] = build_cmd

    # Cache management commands
    cache_cmd = subparser.add_parser("cache",
        help="Manage the IVPM package cache")
    cache_subparser = cache_cmd.add_subparsers(dest="cache_cmd")
    cache_subparser.required = True

    cache_init_cmd = cache_subparser.add_parser("init",
        help="Initialize a new cache directory")
    cache_init_cmd.add_argument("cache_dir",
        help="Path to the cache directory to initialize")
    cache_init_cmd.add_argument("-s", "--shared", dest="shared", action="store_true",
        help="Set group inheritance (chmod g+s) for shared cache usage")
    cache_init_cmd.add_argument("-f", "--force", dest="force", action="store_true",
        help="Force reinitialization of existing directory")

    cache_info_cmd = cache_subparser.add_parser("info",
        help="Show cache information (packages, versions, sizes)")
    cache_info_cmd.add_argument("-c", "--cache-dir", dest="cache_dir",
        help="Cache directory (default: $IVPM_CACHE)")
    cache_info_cmd.add_argument("-v", "--verbose", dest="verbose", action="store_true",
        help="Show detailed version information")

    cache_clean_cmd = cache_subparser.add_parser("clean",
        help="Remove old cache entries")
    cache_clean_cmd.add_argument("-c", "--cache-dir", dest="cache_dir",
        help="Cache directory (default: $IVPM_CACHE)")
    cache_clean_cmd.add_argument("-d", "--days", dest="days", type=int, default=7,
        help="Remove entries older than this many days (default: 7)")

    cache_cmd.set_defaults(func=CmdCache())
    subcommands["cache"] = cache_cmd

    pkginfo_cmd = subparser.add_parser("pkg-info",
        help="Collect paths/files for a listed set of packages")
    pkginfo_cmd.add_argument("type", 
            choices=("incdirs", "paths", "libdirs", "libs", "flags"),
            help="Specifies what info to query")
    pkginfo_cmd.add_argument("-k", "--kind",
            help="Specifies qualifiers on the type of info to query")
    pkginfo_cmd.add_argument("pkgs", nargs="+")
    pkginfo_cmd.set_defaults(func=CmdPkgInfo())
    subcommands["pkginfo"] = pkginfo_cmd

    share_cmd = subparser.add_parser("share",
        help="Returns the 'share' directory, which includes cmake files, etc")
    share_cmd.add_argument("path", nargs=argparse.REMAINDER)
    share_cmd.set_defaults(func=CmdShare())
    subcommands["share"] = share_cmd

    clone_cmd = subparser.add_parser("clone",
        help="Create a new workspace from a Git URL or path")
    clone_cmd.add_argument("src", help="Source URL or path to clone")
    clone_cmd.add_argument("-a", "--anonymous", dest="anonymous", action="store_true",
        help="Clone anonymously (HTTPS); default converts to SSH when applicable")
    clone_cmd.add_argument("-b", "--branch", dest="branch",
        help="Target branch; checks out existing or creates new")
    clone_cmd.add_argument("workspace_dir", nargs="?",
        help="Target workspace directory; defaults to basename of src")
    clone_cmd.add_argument("-d", "--dep-set", dest="dep_set",
        help="Dependency set to use for ivpm update")
    clone_cmd.add_argument("--py-uv", dest="py_uv", action="store_true",
        help="Use 'uv' to manage virtual environment")
    clone_cmd.add_argument("--py-pip", dest="py_pip", action="store_true",
        help="Use 'pip' to manage virtual environment")
    clone_cmd.set_defaults(func=CmdClone())
    subcommands["clone"] = clone_cmd

    update_cmd = subparser.add_parser("update",
        help="Fetches packages specified in ivpm.yaml that have not already been loaded")
    update_cmd.set_defaults(func=CmdUpdate())
    update_cmd.add_argument("-p", "--project-dir", dest="project_dir",
        help="Specifies the project directory to use (default: cwd)")
    update_cmd.add_argument("-d", "--dep-set", dest="dep_set", 
        help="Uses dependencies from specified dep-set instead of 'default-dev'")
    update_cmd.add_argument("-j", "--jobs", dest="jobs", type=int, default=None,
        help="Maximum number of parallel package fetches (default: number of CPU cores)")
    update_cmd.add_argument("-a", "--anonymous-git", dest="anonymous", 
        action="store_true",
        help="Clones git repositories in 'anonymous' mode")
    update_cmd.add_argument("--skip-py-install", "--py-skip-install",
        help="Skip installation of Python packages",
        action="store_true")
    update_cmd.add_argument("--force-py-install", "--py-force-install",
        help="Forces a re-install of Python packages",
        action="store_true")
    update_cmd.add_argument("--py-prerls-packages",
        help="Enable installation of pre-release packages",
        action="store_true")
    update_cmd.add_argument("--py-uv",
        help="Use 'uv' to manage virtual environment",
        action="store_true")
    update_cmd.add_argument("--py-pip",
        help="Use 'uv' to manage virtual environment",
        action="store_true")
    subcommands["update"] = update_cmd
#    update_cmd.add_argument("-r", "--requirements", dest="requirements")
    
    init_cmd = subparser.add_parser("init",
        help="Creates an initial ivpm.yaml file")
    init_cmd.set_defaults(func=CmdInit())
    init_cmd.add_argument("-v", "--version", default="0.0.1")
    init_cmd.add_argument("-f", "--force", default=False, action='store_const', const=True)
    init_cmd.add_argument("name")
    subcommands["init"] = init_cmd
    
    git_status_cmd = subparser.add_parser("git-status",
        help="Runs git status on any git packages (Note: deprecated. use 'status' instead)")
    git_status_cmd.set_defaults(func=CmdGitStatus())
    git_status_cmd.add_argument("-p", "-project-dir", dest="project_dir")
    
    git_update_cmd = subparser.add_parser("git-update",
        help="Updates any git packages (Note: deprecated. use 'sync' instead)")
    git_update_cmd.set_defaults(func=CmdGitUpdate())
    git_update_cmd.add_argument("-p", "-project-dir", dest="project_dir")
    
    snapshot_cmd = subparser.add_parser("snapshot",
        help="Creates a snapshot of required packages")
    snapshot_cmd.set_defaults(func=CmdSnapshot())
    snapshot_cmd.add_argument("-p", "-project-dir", dest="project_dir")
    snapshot_cmd.add_argument("-r", "--rls-deps", dest="rls", action="store_true",
        help="Uses release deps from project root instead of dev deps")
    snapshot_cmd.add_argument("snapshot_dir", 
            help="Specifies the directory where the snapshot will be created")
    subcommands["snapshot"] = snapshot_cmd

    sync_cmd = subparser.add_parser("sync",
        help="Synchronizes dependent packages with an upstream source (if available)")
    sync_cmd.set_defaults(func=CmdSync())

    status_cmd = subparser.add_parser("status",
        help="Checks the status of sub-dependencies such as git repositories")
    status_cmd.set_defaults(func=CmdStatus())

    if parser_ext is not None:
        for ext in parser_ext:
            ext(subparser)

    if options_ext is not None:
        for ext in options_ext:
            ext(subcommands)

    return parser

def main(project_dir=None):
    from .pkg_types.pkg_type_rgy import PkgTypeRgy

    # First things first: load any extensions
    import sys
    if sys.version_info < (3, 10):
        from importlib_metadata import entry_points
    else:
        from importlib.metadata import entry_points

    discovered_plugins = entry_points(group='ivpm.ext')
    parser_ext = []
    options_ext = []
    for p in discovered_plugins:
        try:
            mod = p.load()
            if hasattr(mod, "ivpm_subcommand"):
                parser_ext.append(getattr(mod, "ivpm_subcommand"))
            elif hasattr(mod, "ivpm_options"):
                options_ext.append(mod)
            elif hasattr(mod, "ivpm_pkgtype"):
                pkg_types = []
                getattr(mod, "ivpm_pkgtype")(pkg_types)
                for pt in pkg_types:
                    PkgTypeRgy.inst().register(pt[0], pt[1], pt[2] if len(pt) > 2 else "")
        except Exception as e:
            print("Error: caught exception while loading IVPM extension %s (%s)" %(
                p.name,
                str(e)))
            raise e

    parser = get_parser(parser_ext, options_ext)
    
    # Custom parsing to allow trailing workspace dir after options for 'clone'
    args, extras = parser.parse_known_args()
    if getattr(args, 'command', None) == 'clone' and getattr(args, 'workspace_dir', None) is None:
        if len(extras) == 1:
            args.workspace_dir = extras[0]
            extras = []
    if len(extras) != 0:
        print('ivpm: error: unrecognized arguments: ' + ' '.join(extras))
        sys.exit(2)

    # If the user hasn't specified the project directory,
    # set the default
    if not hasattr(args, "project_dir") or getattr(args, "project_dir") is None:
        args.project_dir = project_dir

    args.func(args)

if __name__ == "__main__":
    main()
    



#****************************************************************************
#* init.py for snapshot
#****************************************************************************
import subprocess
import os
import sys

snapshot_dir = os.path.dirname(os.path.abspath(__file__))

print("Note: python executable %s" % sys.executable)

print("Note: Creating Python virtual environment")
cmd = [sys.executable, '-m', 'venv', os.path.join(snapshot_dir, "python")]
status = subprocess.run(cmd)

if status.returncode != 0:
    print("Python venv-creation command failed: %s" % str(cmd))
    sys.exit(1)
    
if os.path.isfile(os.path.join(snapshot_dir, "python_pkgs.txt")):
    print("Note: installing required Python packages")
    python_dir = os.path.join(snapshot_dir, "python")
    
    # Windows venv
    if os.path.isdir(os.path.join(python_dir, "Scripts")):
        venv_python = os.path.join(python_dir, "Scripts", "python")
    elif os.path.isfile(os.path.join(python_dir, "bin", "python")):
        venv_python = os.path.join(python_dir, "bin", "python")
    else:
        print("Error: unknown python virtual-environment structure")
        sys.exit(1)

    cmd = [venv_python, "-m", "pip", "install", "pip", "setuptools", "--upgrade"]        
    status = subprocess.run(cmd)
    
    if status.returncode != 0:
        print("Error: failed to upgrade pip")
        sys.exit(1)

    cwd = os.getcwd()        
    os.chdir(snapshot_dir)
    cmd = [venv_python, "-m", "pip", "install", "-r", 
           os.path.join(snapshot_dir, "python_pkgs.txt")]
    status = subprocess.run(cmd)
    
    if status.returncode != 0:
        print("Error: failed to install packages from snapshot")
        sys.exit(1)
        
else:
    print("Note: no Python packages to install")
    
print("Done setting up Python virtual environment")

    


'''
Created on Jun 8, 2021

@author: mballance
'''
import os
import sys
from ivpm.arg_utils import ensure_have_project_dir
from ..proj_info import ProjInfo
from ..utils import fatal

class CmdStatus(object):
    
    def __init__(self):
        pass
    
    def __call__(self, args):
        
        if args.project_dir is None:
            args.project_dir = os.getcwd()

        proj_info = ProjInfo.mkFromProj(args.project_dir)

        if proj_info is None:
            fatal("Failed to locate IVPM meta-data (eg ivpm.yaml)")
            
        deps_dir = os.path.join(args.project_dir, proj_info.deps_dir)

        # After that check, go ahead and just check directories
        for dir in os.listdir(deps_dir):
            if os.path.isdir(os.path.join(deps_dir, dir, ".git")):
                print("Package: " + dir)
                cwd = os.getcwd()
                os.chdir(deps_dir + "/" + dir)
                status = os.system("git status -s")
                os.chdir(cwd)
            elif dir != "python" and os.path.isdir(os.path.join(deps_dir, dir)):
                print("Note: skipping non-Git package \"" + dir + "\"")
                sys.stdout.flush()



import os
import sys
import dataclasses as dc
import subprocess
from ..proj_info import ProjInfo
from ..utils import fatal

@dc.dataclass
class CmdActivate(object):

    def __call__(self, args):
        if args.project_dir is None:
            # If a default is not provided, use the current directory
#            print("Note: project_dir not specified ; using working directory")
            args.project_dir = os.getcwd()
            
        proj_info = ProjInfo.mkFromProj(args.project_dir)

        if proj_info is None:
            fatal("Failed to locate IVPM meta-data (eg ivpm.yaml)")

        packages_dir = os.path.join(args.project_dir, "packages")
        if not os.path.isdir(packages_dir):
            fatal("No packages directory ; must run ivpm update first")

        python_dir = os.path.join(packages_dir, "python")
        if not os.path.isdir(python_dir):
            fatal("No packages/python directory ; must run ivpm update first")

        activate = os.path.join(python_dir, "bin/activate")

        # TODO: consider non-bash shells and non-Linux platforms
        shell = getattr(os.environ, "SHELL", "bash")
        cmd = None
        if shell.find("bash") != -1:
            cmd = [shell, "-rcfile",  activate]

            if args.c is not None:
                cmd.extend(["-c", args.c])
        elif shell.find("csh") != -1 or shell.find("ksh") != -1:
            cmd = [shell, "-s",  activate + ".csh"]

            if args.c is None:
                cmd.extend(["-i", args.c])

        cmd.extend(args.args)

        env = os.environ.copy()
        env["IVPM_PROJECT"] = args.project_dir
        env["IVPM_PACKAGES"] = os.path.join(args.project_dir, "packages")
        # env["VIRTUAL_ENV_DISABLE_PROMPT"] = "1"

        # PS1 = getattr(env, "PS1", None)
        # print("PS1: %s" % str(PS1))
        # if PS1 is not None:
        #     PS1 = "(ivpm) %s" % PS1
        # else:
        #     PS1 = "\\[\\](ivpm) "
        # env["PS1"] = PS1

        for es in proj_info.env_settings:
            es.apply(env)

        result = subprocess.run(
            cmd,
            env=env)
        sys.exit(result.returncode)




#****************************************************************************
#* project_ops.py
#*
#* Copyright 2018-2024 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on:
#*     Author: 
#*
#****************************************************************************
import os
import json
import dataclasses as dc
from typing import Tuple, List, Optional
from .package import Package, SourceType
from .package_updater import PackageUpdater
from .handlers.package_handler_rgy import PackageHandlerRgy
from .project_ops_info import ProjectUpdateInfo, ProjectBuildInfo
from .update_listener import UpdateListener, DefaultUpdateListener
from .utils import fatal, note, get_venv_python, setup_venv, warning

@dc.dataclass
class ProjectOps(object):
    root_dir : str
    debug : bool = False

    def update(self,
               dep_set : str = None,
               force_py_install : bool = False,
               skip_venv : bool = False,
               args = None,
               listeners: Optional[List[UpdateListener]] = None):
        proj_info, deps_dir, dep_set = self._init(dep_set)
 
        # Ensure that we have a python virtual environment setup
        if not skip_venv:
            if not os.path.isdir(os.path.join(deps_dir, "python")):
                uv_pip = "auto"
                if hasattr(args, "py_uv") and args.py_uv:
                    uv_pip = "uv"
                elif hasattr(args, "py_pip") and args.py_pip:
                    uv_pip = "pip"
                ivpm_python = setup_venv(os.path.join(deps_dir, "python"), uv_pip=uv_pip)
            else:
                note("python virtual environment already exists")
                ivpm_python = get_venv_python(os.path.join(deps_dir, "python"))
            
        print("********************************************************************")
        print("* Processing root package %s" % proj_info.name)
        print("********************************************************************")

        if self.debug:
            for self.dep_set in proj_info.dep_set_m.keys():
                print("DepSet: %s" % self.dep_set)
                for d in proj_info.dep_set_m[self.dep_set].packages.keys():
                    print("  Package: %s" % d)

        ds = self._getDepSet(proj_info, dep_set)

        # If the root dependency set doesn't specify a source
        # for IVPM, auto-load it from PyPi
        if "ivpm" not in ds.packages.keys():
            print("Note: will install IVPM from PyPi")
            ivpm = Package("ivpm")
            ivpm.src_type = SourceType.PyPi
            ds.packages["ivpm"] = ivpm

        pkg_handler = PackageHandlerRgy.inst().mkHandler()
        updater = PackageUpdater(deps_dir, pkg_handler, args=args)
        
        # Add listeners to the updater
        if listeners:
            for listener in listeners:
                updater.update_info.add_listener(listener)
        else:
            # Add default listener for progress output
            updater.update_info.add_listener(DefaultUpdateListener())

        # Prevent an attempt to load the top-level project as a depedency
        updater.all_pkgs[proj_info.name] = None
        pkgs_info = updater.update(ds)

        print("Setup-deps: %s" % str(pkgs_info.setup_deps))

        # Call the handlers to take care of project-level setup work
        update_info = ProjectUpdateInfo(args, deps_dir, force_py_install=force_py_install, skip_venv=skip_venv)
        pkg_handler.update(update_info)

        # Print cache summary after handlers complete
        updater.update_info.print_cache_summary()

        # Finally, write out some meta-data
        ivpm_json = {}
        ivpm_json["dep-set"] = dep_set
        with open(os.path.join(deps_dir, "ivpm.json"), "w") as fp:
            json.dump(ivpm_json, fp)

    def build(self, dep_set : str = None, args = None, debug : bool = False):
        proj_info, deps_dir, dep_set = self._init(dep_set)

        ds = self._getDepSet(proj_info, dep_set)

        pkg_handler = PackageHandlerRgy.inst().mkHandler()
        updater = PackageUpdater(deps_dir, pkg_handler, args=args, load=False)

        # Prevent an attempt to load the top-level project as a depedency
        updater.all_pkgs[proj_info.name] = None
        pkgs_info = updater.update(ds)

        # Now, run the actual build operation
        build_info = ProjectBuildInfo(args, deps_dir, debug=debug)

        pkg_handler.build(build_info)

        pass

    def status(self, dep_set : str = None):
        pass

    def sync(self, dep_set : str = None):
        pass

    def _init(self, dep_set : str = None) -> Tuple['ProjInfo', str, str]:
        from .proj_info import ProjInfo

        proj_info = ProjInfo.mkFromProj(self.root_dir)

        if proj_info is None:
            fatal("Failed to locate IVPM meta-data (eg ivpm.yaml)")
            
        deps_dir = os.path.join(self.root_dir, proj_info.deps_dir)

        ivpm_json = {}

        if os.path.isfile(os.path.join(deps_dir, "ivpm.json")):
            with open(os.path.join(deps_dir, "ivpm.json"), "r") as fp:
                try:
                    ivpm_json = json.load(fp)
                except Exception as e:
                    warning("failed to read ivpm.json: %s" % str(e))

        if "dep-set" in ivpm_json.keys():
            if dep_set is None:
                dep_set = ivpm_json["dep-set"]
            elif dep_set != ivpm_json["dep-set"]:
                fatal("Attempting to update with a different dep-set than previously used")

        return (proj_info, deps_dir, dep_set)
    
    def _getDepSet(self, proj_info, dep_set):
        if dep_set is None:
            if "default-dev" in proj_info.dep_set_m.keys():
                dep_set = "default-dev"
            elif len(proj_info.dep_set_m.keys()) == 1:
                dep_set = list(proj_info.dep_set_m.keys())[0]
            else:
                fatal("No default-dev dep-set and multiple dep-sets present")

        if dep_set not in proj_info.dep_set_m.keys():
            raise Exception("Dep-set %s is not present" % dep_set)
        else:
            ds = proj_info.dep_set_m[dep_set]
        
        return ds





#****************************************************************************
#* package_http.py
#*
#* Copyright 2023 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on:
#*     Author: 
#*
#****************************************************************************
import os
import httpx
import sys
import urllib
import dataclasses as dc
from .package_file import PackageFile
from ..project_ops_info import ProjectUpdateInfo
from ..utils import note
from ..package import SourceType2Ext
from ..cache import Cache

class PackageHttp(PackageFile):

    def update(self, update_info : ProjectUpdateInfo):
        pkg_dir = os.path.join(update_info.deps_dir, self.name)
        self.path = pkg_dir.replace("\\", "/")

        # Report this package for cache statistics
        # HTTP packages are cacheable if cache=True, editable if cache is not True
        is_cacheable = self.cache is True
        is_editable = self.cache is not True  # Could be cached but isn't
        update_info.report_package(cacheable=is_cacheable, editable=is_editable)

        if os.path.isdir(pkg_dir) or os.path.islink(pkg_dir):
            note("Skipping %s, since it is already loaded" % self.name)
        else:
            # Check if caching is enabled
            if self.cache is True:
                return self._update_with_cache(update_info, pkg_dir)
            elif self.cache is False:
                return self._update_no_cache_readonly(update_info, pkg_dir)
            else:
                return self._update_normal(update_info, pkg_dir)
    
    def _get_url_version(self, url: str) -> str:
        """Get version identifier for a URL using HEAD request.
        
        Uses Last-Modified header or ETag as version identifier.
        """
        try:
            response = httpx.head(url, follow_redirects=True, timeout=30)
            
            # Prefer Last-Modified as it's more human-readable
            if "Last-Modified" in response.headers:
                # Convert to a safe directory name
                lm = response.headers["Last-Modified"]
                # Replace problematic characters
                return lm.replace(" ", "_").replace(":", "-").replace(",", "")
            
            # Fall back to ETag
            if "ETag" in response.headers:
                etag = response.headers["ETag"]
                # Clean up ETag (remove quotes and W/ prefix)
                # Strip quotes from both ends
                etag = etag.strip('"').strip("'")
                # Handle W/ prefix
                if etag.startswith("W/"):
                    etag = etag[2:]
                # Strip quotes again in case W/"..." format
                etag = etag.strip('"').strip("'")
                return etag
            
            # Last resort: use URL hash
            import hashlib
            return hashlib.md5(url.encode()).hexdigest()[:16]
        except Exception:
            import hashlib
            return hashlib.md5(url.encode()).hexdigest()[:16]
    
    def _update_with_cache(self, update_info: ProjectUpdateInfo, pkg_dir: str):
        """Update using the cache."""
        note("loading package %s with cache" % self.name)
        
        # Get version from URL metadata
        version = self._get_url_version(self.url)
        
        cache = update_info.cache
        if cache is None:
            cache = Cache()
        
        # If cache is not properly configured, fall back to no cache
        if not cache.is_enabled():
            note("IVPM_CACHE not set - falling back to no-cache mode for %s" % self.name)
            return self._update_no_cache_readonly(update_info, pkg_dir)
        
        # Check if this version is cached
        if cache.has_version(self.name, version):
            note("Cache hit for %s at version %s" % (self.name, version))
            cache.link_to_deps(self.name, version, update_info.deps_dir)
            update_info.report_cache_hit()
            return
        
        # Cache miss - download and unpack
        note("Cache miss for %s - downloading" % self.name)
        update_info.report_cache_miss()
        
        # Download to temp location
        temp_dir = os.path.join(update_info.deps_dir, f".cache_temp_{self.name}")
        if os.path.exists(temp_dir):
            import shutil
            shutil.rmtree(temp_dir)
        
        download_dir = os.path.join(update_info.deps_dir, ".download")
        if not os.path.isdir(download_dir):
            os.makedirs(download_dir)
        
        if self.unpack:
            pkg_path = os.path.join(download_dir, os.path.basename(self.url))
        else:
            pkg_path = temp_dir
        
        self._download_file(self.url, pkg_path)
        
        if self.unpack:
            self._install(pkg_path, temp_dir)
            os.unlink(pkg_path)
        
        # Store in cache and link
        cache.store_version(self.name, version, temp_dir)
        cache.link_to_deps(self.name, version, update_info.deps_dir)
    
    def _update_no_cache_readonly(self, update_info: ProjectUpdateInfo, pkg_dir: str):
        """Download and make read-only (cache=False)."""
        note("loading package %s (no cache, read-only)" % self.name)
        
        self._update_normal(update_info, pkg_dir)
        
        # Make read-only
        self._make_readonly(pkg_dir)
    
    def _update_normal(self, update_info: ProjectUpdateInfo, pkg_dir: str):
        """Normal download without caching."""
        # Need to fetch, then unpack these
        download_dir = os.path.join(update_info.deps_dir, ".download")
            
        if not os.path.isdir(download_dir):
            os.makedirs(download_dir)

        if self.unpack:
            pkg_path = os.path.join(download_dir, 
                                    os.path.basename(self.url))
        else:
            pkg_path = os.path.join(update_info.deps_dir, self.name)
                
        # TODO: should this be an option?   
        remove_pkg_src = True

        self._download_file(self.url, pkg_path)

        if self.unpack:
            self._install(pkg_path, pkg_dir)
            os.unlink(os.path.join(download_dir, 
                                   os.path.basename(self.url)))
        else:
            # 
            pass
    
    def _make_readonly(self, path: str):
        """Make all files in a directory tree read-only."""
        import stat
        for root, dirs, files in os.walk(path):
            for d in dirs:
                dir_path = os.path.join(root, d)
                mode = os.stat(dir_path).st_mode
                os.chmod(dir_path, mode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)
            for f in files:
                file_path = os.path.join(root, f)
                mode = os.stat(file_path).st_mode
                os.chmod(file_path, mode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)
        mode = os.stat(path).st_mode
        os.chmod(path, mode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)

    def _download_file(self, url, dest):
        r = httpx.get(url, follow_redirects=True)
        with open(dest, "wb") as f:
            f.write(r.content)
        pass
            
    @staticmethod
    def create(name, opts, si) -> 'PackageHttp':
        pkg = PackageHttp(name)
        pkg.process_options(opts, si)
        return pkg








import os
import shutil
import subprocess
import tempfile
import unittest

from .test_base import TestBase

ROOTDIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
SRCDIR = os.path.join(ROOTDIR, 'src')
ENV = os.environ.copy()
ENV['PYTHONPATH'] = SRCDIR + (':' + ENV['PYTHONPATH'] if 'PYTHONPATH' in ENV else '')


class TestClone(TestBase):

    def _init_git_repo(self, path, files=None, branch='main'):
        os.makedirs(path, exist_ok=True)
        subprocess.check_call(["git", "init", "-b", branch], cwd=path, env=ENV)
        # Configure user for committing in this repo
        subprocess.check_call(["git", "config", "user.email", "you@example.com"], cwd=path, env=ENV)
        subprocess.check_call(["git", "config", "user.name", "Your Name"], cwd=path, env=ENV)
        if files:
            for rel, content in files.items():
                full = os.path.join(path, rel)
                os.makedirs(os.path.dirname(full), exist_ok=True)
                with open(full, 'w') as f:
                    f.write(content)
        # Minimal ivpm project to allow update
        if not os.path.exists(os.path.join(path, 'ivpm.yaml')):
            with open(os.path.join(path, 'ivpm.yaml'), 'w') as f:
                f.write('''\npackage:\n  name: sample\n\n  dep-sets:\n    - name: default-dev\n      deps: []\n''')
        subprocess.check_call(["git", "add", "-A"], cwd=path, env=ENV)
        subprocess.check_call(["git", "commit", "-m", "init"], cwd=path, env=ENV)

    @unittest.skip("CI path issues")
    def test_clone_local_git_default(self):
        src_repo = os.path.join(self.testdir, 'src_repo')
        self._init_git_repo(src_repo)

        # Specify a workspace directory different from source to avoid collision
        wsdir = os.path.join(self.testdir, 'workspace_default')
        subprocess.check_call(["python", "-m", "ivpm", "clone", src_repo, wsdir], cwd=self.testdir, env=ENV)
        self.assertTrue(os.path.isdir(wsdir))
        self.assertTrue(os.path.isfile(os.path.join(wsdir, 'packages', 'ivpm.json')))

    @unittest.skip("CI path issues")
    def test_clone_local_git_branch_create(self):
        src_repo = os.path.join(self.testdir, 'src_repo2')
        self._init_git_repo(src_repo)

        wsdir = os.path.join(self.testdir, 'workspace2')
        subprocess.check_call(["python", "-m", "ivpm", "clone", src_repo, wsdir, "-b", "feature/x"], cwd=self.testdir, env=ENV)
        # Verify branch created
        out = subprocess.check_output(["git", "branch", "--show-current"], cwd=wsdir).decode().strip()
        self.assertEqual(out, "feature/x")

    @unittest.skip("CI path issues")
    def test_clone_local_git_branch_checkout_existing(self):
        src_repo = os.path.join(self.testdir, 'src_repo3')
        self._init_git_repo(src_repo)
        # Create a branch and push-like refs locally by creating the branch in src
        subprocess.check_call(["git", "checkout", "-b", "dev"], cwd=src_repo, env=ENV)
        with open(os.path.join(src_repo, 'README.md'), 'w') as f:
            f.write("dev branch")
        subprocess.check_call(["git", "add", "README.md"], cwd=src_repo, env=ENV)
        subprocess.check_call(["git", "commit", "-m", "dev"], cwd=src_repo, env=ENV)
        # Clone and specify branch dev; since there's no remote, origin/dev won't exist. Our logic checks remote via ls-remote which will fail; fallback will create new local branch.
        wsdir = os.path.join(self.testdir, 'workspace3')
        subprocess.check_call(["python", "-m", "ivpm", "clone", src_repo, wsdir, "-b", "dev"], cwd=self.testdir, env=ENV)
        out = subprocess.check_output(["git", "branch", "--show-current"], cwd=wsdir).decode().strip()
        self.assertEqual(out, "dev")



#****************************************************************************
#* setup.py
#*
#* Copyright 2022 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on:
#*     Author: 
#*
#****************************************************************************
import importlib
import os
import sys
from enum import Enum, auto
from setuptools import setup as _setup
from .build_ext import BuildExt
from .install_lib import InstallLib
import inspect
import platform
from ivpm.pkg_info.pkg_info_rgy import PkgInfoRgy
import ivpm.setup.ivpm_data as idata


def setup(*args, **kwargs):
    global _ivpm_extra_data, _ivpm_extdep_data, _ivpm_hooks, _ivpm_ext_name_m

    print("IVPM setup: %s" % kwargs["name"])

    stack = inspect.stack()
    caller = stack[1][0]

    if "-DDEBUG" in sys.argv:
        debug = True
        sys.argv.remove("-DDEBUG")
    else:
        debug = False

    if "ivpm_extra_data" in kwargs.keys():
        idata._ivpm_extra_data = kwargs["ivpm_extra_data"]
        kwargs.pop("ivpm_extra_data")

    if "ivpm_extdep_data" in kwargs.keys():
        idata._ivpm_extdep_data = kwargs["ivpm_extdep_data"]
        kwargs.pop("ivpm_extdep_data")

    if "ivpm_hooks" in kwargs.keys():
        idata._ivpm_hooks = kwargs["ivpm_hooks"]
        kwargs.pop("ivpm_hooks")

    if "ivpm_ext_name_m" in kwargs.keys():
        idata._ivpm_ext_name_m = kwargs["ivpm_ext_name_m"]
        kwargs.pop("ivpm_ext_name_m") 

    # project_dir = os.path.dirname(os.path.abspath(
    #     inspect.getmodule(caller).__file__))

    # if os.path.isdir(os.path.join(project_dir, 'packages')):
    #     packages_dir = os.path.join(project_dir, 'packages')
    # else:
    #     packages_dir = os.path.dirname(os.path.join(project_dir))

    #     if not os.path.isdir(packages_dir):
    #         raise Exception("Failed to locate packages directory: project_dir=%s ; packages_dir=%s" % (
    #             project_dir, packages_dir
    #         ))

    # Update extension flags based on common requirements
    if "ext_modules" in kwargs.keys():
        for m in kwargs["ext_modules"]:
            if hasattr(m, "language") and m.language == "c++":
                print("C++ extension")
                if platform.system() == "Darwin":
                    if not hasattr(m, "extra_compile_args"):
                        setattr(m, "extra_compile_args", [])
                    m.extra_compile_args.append("-std=c++17")
            else:
                print("No 'language': %s" % getattr(m, "language", "<notpresent>"))

    if "ivpm_extdep_pkgs" in kwargs.keys():
        include_dirs = []
        library_dirs = []
        libraries = []
        paths = []
        for dep in kwargs["ivpm_extdep_pkgs"]:
            processed = set()
            _collect_extdeps(
                dep,
                processed,
                include_dirs,
                library_dirs,
                libraries,
                paths)
        kwargs.pop("ivpm_extdep_pkgs")

        print("paths: %s" % str(paths), flush=True)
        print("include_dirs: %s" % str(include_dirs), flush=True)
#        sys.path.extend(paths)

        if "ext_modules" in kwargs.keys():
            for m in kwargs["ext_modules"]:
                # Configure include paths, libraries, etc
                print("Applying extension updates to: %s" % m.name)
                _apply_extdeps(
                    m,
                    include_dirs,
                    library_dirs,
                    libraries)
                print("Final settings for %s:" % m.name)
                print("   incdirs: %s" % str(m.include_dirs))
        else:
            print("Note: no extension libraries")

#    if "BUILD_NUM" in os.environ.keys() and "version" in kwargs:
#        kwargs["version"] += ".%s" % os.environ["BUILD_NUM"]
    
    if "cmdclass" in kwargs:
        cmdclass = kwargs["cmdclass"]
    else:
        cmdclass = {}
        kwargs["cmdclass"] = cmdclass
    
    if "build_ext" in cmdclass.keys():
        print("Warning: build_ext is overridden")
    else:
        cmdclass["build_ext"] = BuildExt
    
    if "install_lib" in cmdclass.keys():
        print("Warning: install_lib is overridden")
    else:
        cmdclass["install_lib"] = InstallLib
    
    print("ivpm.build.setup")
    if "ext_modules" in kwargs:
        print("ext_modules")
        for ext in kwargs["ext_modules"]:
            if hasattr(ext, "package_deps"):
                print("package_deps")
   
    for hook in idata.get_hooks(idata.Phase_SetupPre):
        hook(None)

    _setup(*args, **kwargs)

    for hook in idata.get_hooks(idata.Phase_SetupPost):
        hook(None)

def _collect_extdeps(
    dep,
    processed,
    include_dirs,
    library_dirs,
    libraries,
    paths):
    from ivpm.pkg_info.pkg_info_rgy import PkgInfoRgy
    if dep in processed:
        return
    else:
        processed.add(dep)
    
    rgy = PkgInfoRgy.inst()

    for pkg in rgy.getPkgs():
        print("Package: %s" % pkg, flush=True)

    if rgy.hasPkg(dep):
        print("Package %s is an IVPM package" % dep)
        pkg = rgy.getPkg(dep)

        path = pkg.getPath()
        if path is not None and path not in paths:
            paths.append(path)

        for incdir in pkg.getIncDirs():
            if incdir not in include_dirs:
                include_dirs.append(incdir)

        for libdir in pkg.getLibDirs():
            if libdir not in library_dirs:
                library_dirs.append(libdir)

        for lib in pkg.getLibs():
            if lib not in libraries:
                libraries.append(lib)

        for sub_dep in pkg.getDeps():
            _collect_extdeps(
                sub_dep, 
                processed,
                include_dirs,
                library_dirs,
                libraries,
                paths)
    else:
        # Dep is not an IVPM package
        print("TODO: %s is not an IVPM package" % str(dep))
        try:
            mod = importlib.import_module(dep)
            pkg_path = mod.__file__

            if pkg_path is not None:
                if os.path.isfile(pkg_path):
                    pkg_dir = os.path.dirname(pkg_path)
                else:
                    pkg_dir = pkg_path
                print("pkg_path: %s ; pkg_dir: %s" % (pkg_path, pkg_dir))
                if pkg_dir not in include_dirs:
                    include_dirs.append(pkg_dir)
            else:
                print("Package %s does not have a non-null module path" % dep)
        except ImportError as e:
            print("Failed to import dependency %s (%s)" % (dep, str(e)))


def _apply_extdeps(
    m,
    include_dirs,
    library_dirs,
    libraries):
    for incdir in include_dirs:
        if incdir not in m.include_dirs:
            print("Add incdir %s" % incdir)
            m.include_dirs.append(incdir)
    for libdir in library_dirs:
        if libdir not in m.library_dirs:
            m.library_dirs.append(libdir)
#    for lib in libraries:
#        if lib not in m.libraries:
#            m.libraries.append(lib)



#****************************************************************************
#* IVPM __init__.py
#*
#* Copyright 2018-2023 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on:
#*     Author: 
#*
#****************************************************************************
import os
#from .ivpm_subprocess import ivpm_popen
from .pkg_info.pkg_compile_flags import PkgCompileFlags
from .pkg_info.pkg_info import PkgInfo
from .pkg_info.pkg_info_rgy import PkgInfoRgy
from .utils import load_project_package_info
import ivpm.setup

from .package import Package
from .project_ops_info import ProjectUpdateInfo
from .update_listener import UpdateListener, DefaultUpdateListener, PackageUpdateEvent

def get_pkg_version(setup_py_path):
    """Returns the package version based on the etc/ivpm.info file"""
    rootdir = os.path.dirname(os.path.realpath(setup_py_path))

    version=None
    with open(os.path.join(rootdir, "etc", "ivpm.info"), "r") as fp:
        while True:
            l = fp.readline()
            if l == "":
                break
            if l.find("version=") != -1:
                version=l[l.find("=")+1:].strip()
                break

    if version is None:
        raise Exception("Failed to find version in ivpm.info")

    if "BUILD_NUM" in os.environ.keys():
        version += "." + os.environ["BUILD_NUM"]

    return version

def get_pkg_info(name):
    from .pkg_info.pkg_info_loader import PkgInfoLoader
    if isinstance(name, list):
        return PkgInfoLoader().load_pkgs(name)
    else:
        return PkgInfoLoader().load(name)




#****************************************************************************
#* project_opts_info.py
#*
#* Copyright 2018-2024 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on:
#*     Author: 
#*
#****************************************************************************
import dataclasses as dc
import enum
from typing import Optional, List, TYPE_CHECKING

if TYPE_CHECKING:
    from .update_listener import UpdateListener, PackageUpdateEvent

class FileStatus(enum.Enum):
    Unknown = "U"
    Modified = "M"
    Added = "A"
    Deleted = "D"

@dc.dataclass
class ProjectOpsInfo(object):
    args : object
    deps_dir : str

@dc.dataclass
class ProjectBuildInfo(ProjectOpsInfo):
    debug : bool = False
    pass

@dc.dataclass
class ProjectSyncInfo(ProjectOpsInfo):
    pass

@dc.dataclass
class ProjectStatusInfo(ProjectOpsInfo):
    pass

@dc.dataclass
class ProjectStatusResult(object):
    package : str
    path : str
    status : FileStatus

@dc.dataclass
class ProjectUpdateInfo(ProjectOpsInfo):
    force_py_install : bool = False
    skip_venv : bool = False
    cache: Optional['Cache'] = None
    cache_hits: int = 0
    cache_misses: int = 0
    total_packages: int = 0
    cacheable_packages: int = 0
    editable_packages: int = 0
    max_parallel: int = 0  # 0 means use available cores
    listeners: List['UpdateListener'] = dc.field(default_factory=list)
    _current_event: Optional['PackageUpdateEvent'] = dc.field(default=None, repr=False)
    
    def add_listener(self, listener: 'UpdateListener'):
        """Add a listener to receive update events."""
        self.listeners.append(listener)
    
    def remove_listener(self, listener: 'UpdateListener'):
        """Remove a listener."""
        self.listeners.remove(listener)
    
    def notify_start(self, event: 'PackageUpdateEvent'):
        """Notify listeners that a package update has started."""
        self._current_event = event
        for listener in self.listeners:
            listener.on_package_start(event)
    
    def notify_finish(self, event: 'PackageUpdateEvent'):
        """Notify listeners that a package update has finished."""
        for listener in self.listeners:
            listener.on_package_finish(event)
        self._current_event = None
    
    def report_cache_hit(self):
        self.cache_hits += 1
        if self._current_event is not None:
            self._current_event.cache_hit = True
    
    def report_cache_miss(self):
        self.cache_misses += 1
        if self._current_event is not None:
            self._current_event.cache_hit = False
    
    def report_package(self, cacheable: bool = False, editable: bool = False):
        """Report a package for statistics.
        
        Args:
            cacheable: True if package has cache=True set
            editable: True if package could be cached but isn't (e.g., git repos, .tar.gz without cache=True)
        """
        self.total_packages += 1
        if cacheable:
            self.cacheable_packages += 1
        if editable:
            self.editable_packages += 1
        if self._current_event is not None:
            self._current_event.is_cacheable = cacheable
            self._current_event.is_editable = editable
    
    def print_cache_summary(self):
        """Print a summary of cache statistics."""
        print("")
        print("Sub-Package Update Summary:")
        print("  Total packages: %d" % self.total_packages)
        print("  Cacheable packages: %d" % self.cacheable_packages)
        print("  Editable packages: %d" % self.editable_packages)
        if self.cacheable_packages > 0:
            print("  Cache hits: %d" % self.cache_hits)
            print("  Cache misses: %d" % self.cache_misses)
            hit_rate = (self.cache_hits / self.cacheable_packages * 100) if self.cacheable_packages > 0 else 0
            print("  Hit rate: %.1f%%" % hit_rate)





#****************************************************************************
#* package.py
#*
#* Copyright 2018-2024 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on: Jun 8, 2021
#*     Author: mballance
#*
#****************************************************************************

import os
import dataclasses as dc
from enum import Enum, auto
from typing import Dict, List, Set
from .project_ops_info import ProjectUpdateInfo
from .utils import fatal, getlocstr

class PackageType(Enum):
    Raw = auto()
    Python = auto()
    Unknown = auto()
    
PackageType2Spec = {
    PackageType.Raw     : "raw",
    PackageType.Python  : "python",
    PackageType.Unknown : "unknown"
    }

Spec2PackageType = {
    "raw"     : PackageType.Raw,
    "python"  : PackageType.Python,
    "unknown" : PackageType.Unknown
    }
    
class SourceType(Enum):
    Git = auto()
    Jar = auto()
    Tgz = auto()
    Txz = auto()
    Zip = auto()
    PyPi = auto()

Ext2SourceType = {
        ".git" : SourceType.Git,
        ".jar" : SourceType.Jar,
        ".tar.gz" : SourceType.Tgz,
        ".tar.xz" : SourceType.Txz,
        ".tgz" : SourceType.Tgz,
        ".zip" : SourceType.Zip
        }

SourceType2Ext = {
        SourceType.Git : ".git",
        SourceType.Jar : ".jar",
        SourceType.Tgz : ".tar.gz",
        SourceType.Txz : ".tar.xz",
        SourceType.Zip : ".zip"
    }

SourceType2Spec = {
        SourceType.Git  : "git",
        SourceType.Jar  : "jar",
        SourceType.Tgz  : "tgz",
        SourceType.Txz  : "txz",
        SourceType.Zip  : "zip",
        SourceType.PyPi : "pypi",
    }
Spec2SourceType = {
        "git"  : SourceType.Git,
        "jar"  : SourceType.Jar,
        "tgz"  : SourceType.Tgz,
        "txz"  : SourceType.Txz,
        "zip"  : SourceType.Zip,
        "pypi" : SourceType.PyPi
    }

@dc.dataclass
class Package(object):
    """Contains leaf-level information about a single package"""
    name : str
    srcinfo : object = None
    path : str = None
    pkg_type : PackageType = None
    src_type : str = None

    process_deps : bool = True
    setup_deps : Set[str] = dc.field(default_factory=set)
    dep_set : str = None
    proj_info : 'ProjInfo'= None

    def build(self, pkgs_info):
        pass

    def status(self, pkgs_info):
        pass

    def sync(self, pkgs_info):
        pass

    def update(self, update_info : ProjectUpdateInfo) -> 'ProjInfo':
        from .proj_info import ProjInfo

        # Report this package for cache statistics (base packages are not cacheable)
        update_info.report_package(cacheable=False)

        info = ProjInfo.mkFromProj(
            os.path.join(update_info.deps_dir, self.name))
        
        return info
    
    def process_options(self, opts, si):
        self.srcinfo = si

        if "dep-set" in opts.keys():
            print("Using dep-set %s for package %s" % (
                opts["dep-set"], self.name))
            self.dep_set = opts["dep-set"]

        if "deps" in opts.keys():
            if opts["deps"] == "skip":
                self.process_deps = False
            else:
                fatal("Unknown value for 'deps': %s" % opts["deps"])

        # Determine the package type (eg Python, Raw)
        if "type" in opts.keys():
            type_s = opts["type"]
            if not type_s in Spec2PackageType.keys():
                fatal("unknown package type %s @ %s ; Supported types 'raw', 'python'" % (
                    type_s, getlocstr(opts["type"])))
                    
            self.pkg_type = Spec2PackageType[type_s]
    
    @staticmethod
    def mk(name, opts, si) -> 'Package':
        raise NotImplementedError()
    





#****************************************************************************
#* package_file.py
#*
#* Copyright 2023 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on:
#*     Author: 
#*
#****************************************************************************
import os
import sys
import shutil
import tarfile
from zipfile import ZipFile
import dataclasses as dc
from .package_url import PackageURL
from ..proj_info import ProjInfo
from ..project_ops_info import ProjectUpdateInfo

@dc.dataclass
class PackageFile(PackageURL):
    unpack : bool = None

    def update(self, update_info : ProjectUpdateInfo) -> ProjInfo:
        # Report this package for cache statistics (file packages are not cacheable by default)
        update_info.report_package(cacheable=False)

        pkg_dir = os.path.join(update_info.deps_dir, self.name)
        self.path = pkg_dir.replace("\\", "/")

        if not os.path.isdir(pkg_dir):
            # TODO: need to fetch the source...

            # Install (unpack) the file 
            if self.unpack:
                self._install(self.url, pkg_dir)

        if os.path.isdir(pkg_dir):
            return ProjInfo.mkFromProj(pkg_dir)
        else:
            return None
    
    def _install(self, pkg_src, pkg_path):
        if self.src_type in (".tar.gz", ".tar.xz", ".tar.bz2"):
            self._install_tgz(pkg_src, pkg_path)
        elif self.src_type in (".jar", ".zip"):
            self._install_zip(pkg_src, pkg_path)
        else:
            raise Exception("Unsupported src_type: %s" % self.src_type)

    def _install_tgz(self, pkg_src, pkg_path):
        cwd = os.getcwd()
        try:
            os.chdir(os.path.dirname(pkg_path))
        
            tf = tarfile.open(pkg_src)

            for fi in tf:
                if fi.name.find("/") != -1:
                    fi.name = fi.name[fi.name.find("/")+1:]
                    tf.extract(fi, path=os.path.basename(pkg_path))
            tf.close()
        finally:
            os.chdir(cwd)

    def _install_zip(self, pkg_src, pkg_path):
            cwd = os.getcwd()
            try:
                os.chdir(os.path.dirname(pkg_path))
                sys.stdout.flush()
                with ZipFile(pkg_src, 'r') as zipObj:
                    zipObj.extractall(os.path.basename(pkg_path))
            finally:
                os.chdir(cwd)        

    def process_options(self, opts, si):
        super().process_options(opts, si)

        if "src" in opts.keys():
            self.src_type = opts["src"]
        else:
            ext = os.path.splitext(self.url)
            if ext == ".tgz":
                self.src_type = ".tar.gz"
            else:
                self.src_type = os.path.splitext(self.url)[1]
                if self.src_type in [".gz", ".xz", ".bz2"]:
                    pdot = self.url.rfind('.')
                    pdot = self.url.rfind('.', 0, pdot-1)
                    self.src_type = self.url[pdot:]

        if "unpack" in opts.keys():
            self.unpack = opts["unpack"]
        else:
            if self.src_type in [".jar"]:
                self.unpack = False
            else:
                self.unpack = True

    @staticmethod
    def create(name, opts, si) -> 'PackageFile':
        pkg = PackageFile(name)
        pkg.process_options(opts, si)
        return pkg




#****************************************************************************
#* package_gh_rls.py
#*
#* Copyright 2023 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on:
#*     Author: 
#*
#****************************************************************************
import os
import httpx
import json
import re
import platform
import subprocess
import shutil
import dataclasses as dc
from typing import Optional
from ..proj_info import ProjInfo
from ..cache import Cache
from ..utils import note
from .package_http import PackageHttp

@dc.dataclass
class PackageGhRls(PackageHttp):
    version : str = "latest"
    file : Optional[str] = None
    prerelease : bool = False  # Whether to include prerelease releases when selecting

    def process_options(self, opts, si):
        super().process_options(opts, si)

        if self.url.find("github.com") == -1:
            raise Exception("GitHub release URL must be specified. URL %s doesn't contain github.com" % self.url)
        
        if "version" in opts.keys():
            self.version = opts["version"]

        if "file" in opts.keys():
            self.file = opts["file"]

        if "prerelease" in opts.keys():
            self.prerelease = bool(opts["prerelease"])


    def update(self, update_info):
        # Report this package for cache statistics
        # GitHub release packages are cacheable if cache=True, editable if cache is not True
        is_cacheable = self.cache is True
        is_editable = self.cache is not True  # Could be cached but isn't
        update_info.report_package(cacheable=is_cacheable, editable=is_editable)

        pkg_dir = os.path.join(update_info.deps_dir, self.name)

        if os.path.isdir(pkg_dir) or os.path.islink(pkg_dir):
            note("Skipping %s, since it is already loaded" % self.name)
            return

        # Query release metadata
        rls_info, rls, file_url, forced_ext = self._resolve_release()

        # Get version from release tag for caching
        release_tag = rls.get("tag_name", "")

        # Check if caching is enabled
        if self.cache is True:
            return self._update_with_cache(update_info, pkg_dir, file_url, forced_ext, release_tag)
        elif self.cache is False:
            return self._update_no_cache_readonly(update_info, pkg_dir, file_url, forced_ext)
        else:
            return self._update_normal(update_info, pkg_dir, file_url, forced_ext)

    def _resolve_release(self):
        """Query GitHub API and resolve the release and asset to download.
        
        Returns:
            Tuple of (rls_info, rls, file_url, forced_ext)
        """
        github_com_idx = self.url.find("github.com")
        url = "https://api.github.com/repos/" + self.url[github_com_idx + len("github.com") + 1:] + "/releases"
        rls_info = httpx.get(url, follow_redirects=True)

        if rls_info.status_code != 200:
            raise Exception("Failed to fetch release info: %d" % rls_info.status_code)

        rls_info = json.loads(rls_info.content)

        # Select release per version specification
        rls = None
        if self.version == "latest":
            for r in rls_info:
                if r["prerelease"] and not self.prerelease:
                    continue
                rls = r
                break
            if rls is None:
                raise Exception("Failed to find latest release (prerelease=%s)" % self.prerelease)
        else:
            rls = self._select_release_by_version(rls_info)
            if rls is None:
                raise Exception(f"No release matches version spec '{self.version}' (prerelease={self.prerelease})")

        # Determine file to download
        file_url = None
        forced_ext = None

        assets = rls.get("assets", [])
        if self.file is not None:
            raise NotImplementedError("File specification not yet supported")

        has_binaries = self._has_binary_assets(assets)
        if not has_binaries:
            file_url, forced_ext = self._choose_source_url(rls)
            if file_url is None:
                if len(assets) == 1 and "browser_download_url" in assets[0]:
                    file_url = assets[0]["browser_download_url"]
                else:
                    raise Exception("No suitable source artifact found")
        else:
            sysname, machine, glibc = self._get_system_info()
            norm_arch = self._normalize_arch(sysname, machine)

            selected = None
            if sysname == "linux":
                if glibc is None:
                    raise Exception("Cannot determine glibc version on Linux.")
                selected = self._select_linux_asset(assets, norm_arch, glibc)
                if selected is None:
                    found = []
                    for a in assets:
                        nm = (a.get("name") or os.path.basename(a.get("browser_download_url", ""))).lower()
                        tag = self._parse_manylinux(nm)
                        if tag is not None:
                            found.append(f"manylinux_{tag[0]}_{tag[1]}_{tag[2]}")
                    if not found:
                        names = [a.get("name") or os.path.basename(a.get("browser_download_url", "")) for a in assets]
                        raise Exception(f"No suitable Linux binary found for arch={norm_arch} glibc={glibc[0]}.{glibc[1]}. Available assets: {', '.join(names)}")
                    else:
                        raise Exception(f"No suitable Linux manylinux binary found for arch={norm_arch} glibc={glibc[0]}.{glibc[1]}. Available manylinux tags: {', '.join(sorted(set(found)))}")
            elif sysname == "darwin":
                selected = self._select_macos_asset(assets, norm_arch)
                if selected is None:
                    names = [a.get("name") or os.path.basename(a.get("browser_download_url", "")) for a in assets]
                    raise Exception(f"No suitable macOS binary found for arch={norm_arch}. Available assets: {', '.join(names)}")
            elif sysname == "windows":
                selected = self._select_windows_asset(assets, norm_arch)
                if selected is None:
                    names = [a.get("name") or os.path.basename(a.get("browser_download_url", "")) for a in assets]
                    raise Exception(f"No suitable Windows binary found for arch={norm_arch}. Available assets: {', '.join(names)}")
            else:
                raise Exception(f"Unsupported platform: {sysname}")

            file_url = selected["browser_download_url"]

        return rls_info, rls, file_url, forced_ext

    def _determine_src_type(self, file_url, forced_ext):
        """Determine the source type for unpacking."""
        if forced_ext is not None:
            ext = forced_ext
        else:
            ext = os.path.splitext(file_url)[1]
            if ext == "":
                ext = ".tar.gz"

        if ext == ".tgz":
            self.src_type = ".tar.gz"
        else:
            self.src_type = ext
            if self.src_type in [".gz", ".xz", ".bz2"]:
                pdot = file_url.rfind('.')
                pdot = file_url.rfind('.', 0, pdot - 1)
                self.src_type = file_url[pdot:]

    def _update_with_cache(self, update_info, pkg_dir, file_url, forced_ext, release_tag):
        """Update using the cache."""
        note("loading package %s with cache" % self.name)

        # Use release tag as version identifier for caching
        # Include platform info for binary releases to cache per-platform
        sysname, machine, _ = self._get_system_info()
        norm_arch = self._normalize_arch(sysname, machine)
        version = f"{release_tag}_{sysname}_{norm_arch}"

        cache = update_info.cache
        if cache is None:
            cache = Cache()

        if not cache.is_enabled():
            note("IVPM_CACHE not set - falling back to no-cache mode for %s" % self.name)
            return self._update_no_cache_readonly(update_info, pkg_dir, file_url, forced_ext)

        if cache.has_version(self.name, version):
            note("Cache hit for %s at version %s" % (self.name, version))
            cache.link_to_deps(self.name, version, update_info.deps_dir)
            update_info.report_cache_hit()
            return

        note("Cache miss for %s - downloading" % self.name)
        update_info.report_cache_miss()

        # Download to temp location
        temp_dir = os.path.join(update_info.deps_dir, f".cache_temp_{self.name}")
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)

        self._determine_src_type(file_url, forced_ext)
        filename = os.path.basename(file_url)
        if forced_ext is not None and not filename.endswith(forced_ext):
            filename = filename + forced_ext
        download_dst = os.path.join(update_info.deps_dir, filename)
        self._download_file(file_url, download_dst)

        self._install(download_dst, temp_dir)
        os.unlink(download_dst)

        cache.store_version(self.name, version, temp_dir)
        cache.link_to_deps(self.name, version, update_info.deps_dir)

    def _update_no_cache_readonly(self, update_info, pkg_dir, file_url, forced_ext):
        """Download and make read-only (cache=False)."""
        note("loading package %s (no cache, read-only)" % self.name)

        self._update_normal(update_info, pkg_dir, file_url, forced_ext)
        self._make_readonly(pkg_dir)

    def _update_normal(self, update_info, pkg_dir, file_url, forced_ext):
        """Normal download without caching."""
        self._determine_src_type(file_url, forced_ext)

        filename = os.path.basename(file_url)
        if forced_ext is not None and not filename.endswith(forced_ext):
            filename = filename + forced_ext
        download_dst = os.path.join(update_info.deps_dir, filename)
        self._download_file(file_url, download_dst)

        self._install(download_dst, pkg_dir)
        os.unlink(download_dst)

    def _parse_version_tuple(self, v):
        m = re.match(r'v?(\d+(?:\.\d+)*)', v or '')
        if not m:
            return None
        return tuple(int(p) for p in m.group(1).split('.'))

    def _cmp_versions(self, a, b):
        la = len(a); lb = len(b)
        for i in range(max(la, lb)):
            ai = a[i] if i < la else 0
            bi = b[i] if i < lb else 0
            if ai != bi:
                return 1 if ai > bi else -1
        return 0

    def _select_release_by_version(self, releases):
        spec = self.version.strip()
        m = re.match(r'(>=|<=|>|<)\s*v?(\d+(?:\.\d+)*)$', spec)
        if m:
            op = m.group(1)
            tgt = self._parse_version_tuple(m.group(2))
            best = None
            for r in releases:
                if r["prerelease"] and not self.prerelease:
                    continue
                rv = self._parse_version_tuple(r.get("tag_name",""))
                if rv is None:
                    continue
                c = self._cmp_versions(rv, tgt)
                ok = False
                if op == ">": ok = c > 0
                elif op == ">=": ok = c >= 0
                elif op == "<": ok = c < 0
                elif op == "<=": ok = c <= 0
                if not ok:
                    continue
                if op in (">", ">="):
                    best = r
                    break
                else: # < or <= : keep highest satisfying
                    if best is None:
                        best = r
                    else:
                        b_v = self._parse_version_tuple(best.get("tag_name",""))
                        if self._cmp_versions(rv, b_v) > 0:
                            best = r
            return best
        # Exact match (allow optional leading v)
        for r in releases:
            if r["prerelease"] and not self.prerelease:
                continue
            tag = r.get("tag_name","")
            if tag == spec or tag == f"v{spec}" or spec == tag.lstrip('v'):
                return r
        tgt = self._parse_version_tuple(spec)
        if tgt:
            for r in releases:
                if r["prerelease"] and not self.prerelease:
                    continue
                rv = self._parse_version_tuple(r.get("tag_name",""))
                if rv == tgt:
                    return r
        return None

    # Helper methods for asset selection and platform detection

    def _normalize_arch(self, system, machine):
        system = (system or "").lower()
        m = (machine or "").lower()
        if system == "linux":
            if m in ("x86_64", "amd64"):
                return "x86_64"
            if m in ("aarch64", "arm64"):
                return "aarch64"
            if m.startswith("armv7"):
                return "armv7l"
            return m
        elif system == "darwin":
            if m in ("arm64", "aarch64"):
                return "arm64"
            return "x86_64" if m in ("x86_64", "amd64") else m
        elif system == "windows":
            return "x86_64" if m in ("x86_64", "amd64") else m
        else:
            return m

    def _get_system_info(self):
        sysname = platform.system().lower()
        machine = platform.machine()
        glibc = None
        if sysname == "linux":
            libc_name, libc_ver = platform.libc_ver()
            ver_t = None
            if libc_ver:
                try:
                    parts = libc_ver.split(".")
                    major = int(parts[0]) if len(parts) > 0 else 0
                    minor = int(parts[1]) if len(parts) > 1 else 0
                    ver_t = (major, minor)
                except Exception:
                    ver_t = None
            if ver_t is None:
                try:
                    # Fallback to parsing ldd --version
                    out = subprocess.check_output(["ldd", "--version"], stderr=subprocess.STDOUT, text=True)
                    m = re.search(r"(\d+)\.(\d+)", out)
                    if m:
                        ver_t = (int(m.group(1)), int(m.group(2)))
                except Exception:
                    ver_t = None
            glibc = ver_t
        return sysname, machine, glibc

    def _parse_manylinux(self, name):
        """
        Parse manylinux tag from a filename.
        Returns (glibc_major, glibc_minor, arch) or None
        """
        n = (name or "").lower()
        m = re.search(r"manylinux_(\d+)_(\d+)_([a-z0-9_]+)", n)
        if m:
            return (int(m.group(1)), int(m.group(2)), m.group(3))
        m = re.search(r"manylinux(1|2010|2014)_([a-z0-9_]+)", n)
        if m:
            legacy = m.group(1)
            arch = m.group(2)
            if legacy == "1":
                return (2, 5, arch)       # manylinux1 -> glibc 2.5
            elif legacy == "2010":
                return (2, 12, arch)      # manylinux2010 -> glibc 2.12
            elif legacy == "2014":
                return (2, 17, arch)      # manylinux2014 -> glibc 2.17
        return None

    def _select_linux_asset(self, assets, arch, glibc):
        candidates = []
        for a in assets:
            nm = (a.get("name") or os.path.basename(a.get("browser_download_url", ""))).lower()
            tag = self._parse_manylinux(nm)
            if tag is None:
                continue
            if tag[2] != arch:
                continue
            ver = (tag[0], tag[1])
            candidates.append((ver, a))
        if not candidates:
            return None
        eligible = [(ver, a) for (ver, a) in candidates if ver <= glibc]
        if eligible:
            eligible.sort(key=lambda x: x[0])
            return eligible[-1][1]
        return None

    def _select_macos_asset(self, assets, arch):
        # Minimal heuristic: look for mac-related tags and arch token
        keys = ("macos", "darwin", "osx")
        for a in assets:
            nm = (a.get("name") or os.path.basename(a.get("browser_download_url", ""))).lower()
            if any(k in nm for k in keys):
                if arch in nm or (arch == "arm64" and "aarch64" in nm) or (arch == "x86_64" and "amd64" in nm):
                    return a
        # Fallback: any mac-related asset
        for a in assets:
            nm = (a.get("name") or os.path.basename(a.get("browser_download_url", ""))).lower()
            if any(k in nm for k in keys):
                return a
        return None

    def _select_windows_asset(self, assets, arch):
        keys = ("windows", "win64", "win32", "win")
        arch_aliases = [arch]
        if arch == "x86_64":
            arch_aliases += ["amd64"]
        for a in assets:
            nm = (a.get("name") or os.path.basename(a.get("browser_download_url", ""))).lower()
            if any(k in nm for k in keys) and any(alias in nm for alias in arch_aliases):
                return a
        # Fallback to any windows asset
        for a in assets:
            nm = (a.get("name") or os.path.basename(a.get("browser_download_url", ""))).lower()
            if any(k in nm for k in keys):
                return a
        return None

    def _has_binary_assets(self, assets):
        for a in assets:
            nm = (a.get("name") or os.path.basename(a.get("browser_download_url", ""))).lower()
            if self._parse_manylinux(nm) is not None:
                return True
            if any(k in nm for k in ("macos", "darwin", "osx", "windows", "win64", "win32", "win")):
                return True
        return False

    def _choose_source_url(self, rls):
        # Prefer tarball over zipball
        tarball = rls.get("tarball_url")
        if tarball:
            return tarball, ".tar.gz"
        zipball = rls.get("zipball_url")
        if zipball:
            return zipball, ".zip"
        return None, None

    @staticmethod
    def create(name, opts, si) -> 'PackageGhRls':
        pkg = PackageGhRls(name)
        pkg.process_options(opts, si)
        return pkg



import os
import platform
import unittest

from .test_base import TestBase


def _check_github_api_available():
    """Check if GitHub API is accessible (not rate-limited)."""
    try:
        import httpx
        resp = httpx.get("https://api.github.com/rate_limit", timeout=5)
        if resp.status_code == 403:
            return False
        return True
    except Exception:
        return False


class TestGhRls(TestBase):

    @unittest.skipUnless(platform.system().lower() == "linux", "Linux-only test")
    @unittest.skipUnless(_check_github_api_available(), "GitHub API rate-limited or unavailable")
    def test_verilator_bin_linux(self):
        # Create a project that depends on a GitHub Release package
        self.mkFile("ivpm.yaml", """
        package:
            name: gh_rls_verilator
            dep-sets:
                - name: default-dev
                  deps:
                    - name: verilator-bin
                      url: https://github.com/pss-hands-on/verilator-bin
                      src: gh-rls
        """)

        # Fetch and install dependencies (no venv needed)
        self.ivpm_update(skip_venv=True)

        pkg_dir = os.path.join(self.testdir, "packages", "verilator-bin")
        self.assertTrue(os.path.isdir(pkg_dir), "verilator-bin package directory missing")

        # Locate the 'verilator' executable in the unpacked package
        exe_path = None
        for root, _, files in os.walk(pkg_dir):
            if "verilator" in files:
                candidate = os.path.join(root, "verilator")
                if os.path.isfile(candidate) and os.access(candidate, os.X_OK):
                    exe_path = candidate
                    break

        self.assertIsNotNone(exe_path, "verilator executable not found in installed package")

        # Running '--version' verifies we selected a runnable Linux binary
        out = self.exec([exe_path, "--version"], cwd=self.testdir)
        self.assertIn("Verilator", out)



'''
Created on Jun 22, 2021

@author: mballance
'''
import sys


def note(msg):
    print("Note: %s" % msg)
    sys.stdout.flush()

def warning(msg):
    print("Warning: %s" % msg)
    sys.stdout.flush()
        
def error(msg):
    print("Error: %s" % msg)
    sys.stdout.flush()
    
def fatal(msg):
    print("Fatal: %s" % msg)
    raise Exception(msg)



#****************************************************************************
#* project_updater.py
#*
#* Copyright 2018-2024 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on: Jun 22, 2021
#*     Author: mballance
#*
#****************************************************************************
import asyncio
import os
import shutil
import subprocess
import sys
import tarfile
import urllib
from zipfile import ZipFile

from ivpm.msg import note, fatal, warning
from ivpm.package import Package, SourceType, SourceType2Ext, PackageType
from ivpm.packages_info import PackagesInfo
from ivpm.proj_info import ProjInfo
from ivpm.update_listener import PackageUpdateEvent
from typing import Dict, List, Tuple
from ivpm.utils import get_venv_python
from .project_ops_info import ProjectUpdateInfo


class PackageUpdater(object):
    
    def __init__(self, 
                 deps_dir, 
                 pkg_handler,
                 load=True,
                 args=None):
        self.debug = False
        self.deps_dir = deps_dir
        self.pkg_handler = pkg_handler
        self.all_pkgs = PackagesInfo("root")
        self.new_deps = []
        self.args = object() if args is None else args
        self.load = load
        self.update_info = ProjectUpdateInfo(self.args, deps_dir)
        
        # Get max parallelism from args, default to CPU count
        if hasattr(args, 'jobs') and args.jobs is not None:
            self.max_parallel = args.jobs
        else:
            import multiprocessing
            self.max_parallel = multiprocessing.cpu_count()
        self.update_info.max_parallel = self.max_parallel
        pass
    
    def update(self, pkgs : PackagesInfo) -> PackagesInfo:
        """
        Updates the specified packages, handling dependencies.
        Uses async parallel fetching for efficiency.
        """
        return asyncio.run(self._update_async(pkgs))
    
    async def _update_async(self, pkgs: PackagesInfo) -> PackagesInfo:
        """
        Async implementation of update that processes packages in parallel.
        The 'pkgs' parameter holds the dependency information
        from the root project.
        """
        count = 1

        pkg_q = []
        
        if len(pkgs.keys()) == 0:
            print("No packages")

        for key in pkgs.keys():
            print("Package: %s" % key)
            pkg_q.append(pkgs[key])

        if not os.path.isdir(self.deps_dir):
            os.makedirs(self.deps_dir)

        # Create semaphore to limit parallelism
        semaphore = asyncio.Semaphore(self.max_parallel)
        
        while True:        
            pkg_deps = {}
            
            # Process this batch of packages in parallel
            if len(pkg_q) > 0:
                results = await self._process_batch_parallel(pkg_q, semaphore)
                
                # Collect dependencies from results
                for pkg, proj_info in results:
                    self.all_pkgs[pkg.name] = pkg
                    
                    # proj_info contains info on any setup-deps that
                    # might be required
                    if proj_info is not None:
                        for sd in proj_info.setup_deps:
                            print("Add setup-dep %s to package %s" % (sd, pkg.name))
                            if pkg.name not in self.all_pkgs.setup_deps.keys():
                                self.all_pkgs.setup_deps[pkg.name] = set()
                            self.all_pkgs.setup_deps[pkg.name].add(sd)

                        if proj_info.process_deps:
                            if not proj_info.has_dep_set(pkg.dep_set):
                                fatal("package %s in %s does not contain specified dep-set %s" % (
                                    proj_info.name, 
                                    pkg.name,
                                    pkg.dep_set))
                                continue
                            else:
                                note("Loading package %s dependencies from dep-set %s" % (proj_info.name, pkg.dep_set))

                            note("Processing dep-set %s of project %s" % (
                                pkg.dep_set,
                                pkg.name))                        

                            ds : PackagesInfo = proj_info.get_dep_set(pkg.dep_set)
                            for d in ds.packages.keys():
                                dep = ds.packages[d]
                        
                                if dep.name not in pkg_deps.keys():
                                    pkg_deps[dep.name] = dep
                                else:
                                    # TODO: warn about possible version conflict?
                                    pass
            
            # Collect new dependencies and add to queue
            pkg_q = []
            for key in pkg_deps.keys():
                if not key in self.all_pkgs.keys():
                    # New package
                    pkg_q.append(pkg_deps[key])
            note("%d new dependencies from iteration %d" % (len(pkg_q), count))
                    
            if len(pkg_q) == 0:
                # We're done
                break
            
            count += 1
            
        return self.all_pkgs
    
    async def _process_batch_parallel(self, pkg_q: List[Package], semaphore: asyncio.Semaphore) -> List[Tuple[Package, ProjInfo]]:
        """Process a batch of packages in parallel."""
        tasks = []
        for pkg in pkg_q:
            tasks.append(self._update_pkg_async(pkg, semaphore))
        
        results = await asyncio.gather(*tasks, return_exceptions=True)
        
        # Process results, handling any exceptions
        processed = []
        for i, result in enumerate(results):
            pkg = pkg_q[i]
            if isinstance(result, Exception):
                # Notify listeners of failure
                event = PackageUpdateEvent(name=pkg.name, error=str(result))
                self.update_info.notify_finish(event)
                fatal("Failed to update package %s: %s" % (pkg.name, str(result)))
            else:
                processed.append(result)
        
        return processed
    
    async def _update_pkg_async(self, pkg: Package, semaphore: asyncio.Semaphore) -> Tuple[Package, ProjInfo]:
        """Async wrapper for updating a single package with semaphore limiting."""
        async with semaphore:
            return await asyncio.get_event_loop().run_in_executor(
                None, self._update_pkg, pkg
            )
    
    def _update_pkg(self, pkg : Package) -> Tuple[Package, ProjInfo]:
        """Loads a single package. Returns the package and any dependencies."""
        must_update=False

        print("********************************************************************")
        print("* Processing package %s (dep-set %s)" % (pkg.name, pkg.dep_set))
        print("********************************************************************")

        # Create and notify start event
        event = PackageUpdateEvent(name=pkg.name)
        self.update_info.notify_start(event)

        pkg_dir = os.path.join(self.deps_dir, pkg.name)
        pkg.path = pkg_dir.replace("\\", "/")

        try:
            pkg.proj_info = pkg.update(self.update_info)

            # Notify the package handlers after the source is 
            # loaded so they can take further action if required 
            self.pkg_handler.process_pkg(pkg)
            
            # Ensure that we use the requested dep-set
            if pkg.proj_info is not None:
                pkg.proj_info.target_dep_set = pkg.dep_set
                pkg.proj_info.process_deps = pkg.process_deps
            
            # Notify finish with current event state (cache_hit etc. updated by update())
            self.update_info.notify_finish(event)
            
            return (pkg, pkg.proj_info)
        except Exception as e:
            event.error = str(e)
            self.update_info.notify_finish(event)
            raise

    

    




#****************************************************************************
#* proj_info.py
#*
#* Copyright 2018-2024 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on: Jan 19, 2020
#*     Author: mballance
#*
#****************************************************************************
import os
from enum import Enum, auto
from ivpm.packages_info import PackagesInfo
from .ivpm_yaml_reader import IvpmYamlReader
from .msg import error, fatal, note
from typing import Dict, List
from .env_spec import EnvSpec

class ProjInfo():
    """Holds information read about a project from its IVPM file"""
    def __init__(self, is_src):
        self.is_src = is_src
        self.dependencies = []
        self.is_legacy = False

        # Dep-set to use when loading sub-dependencies
        self.target_dep_set = "default"
        self.dep_set_m : Dict[str,PackagesInfo] = {}
        self.setup_deps = set()
        
        self.ivpm_info = {}
        self.requirements_txt = None

        self.name = None
        self.version = None
        # This should be set to the dep-set specified by the 'dep' 
        # statement or on the command-line
        self.default_dep_set = None
        self.deps_dir = "packages"

        self.process_deps = True
        self.paths : Dict[str, Dict[str, List[str]]] = {}
        self.env_settings : List[EnvSpec] = []

    def has_dep_set(self, name):
        return name in self.dep_set_m.keys()
            
    def get_dep_set(self, name):
        return self.dep_set_m[name]
    
    def get_target_dep_set(self):
        if self.target_dep_set is None:
            raise Exception("target_dep_set is not specified")
        if self.target_dep_set not in self.dep_set_m.keys():
            raise Exception("Dep-set %s is not present in project %s" % (
                self.target_dep_set, self.name))
        return self.dep_set_m[self.target_dep_set]
    
    def set_dep_set(self, name, ds):
        self.dep_set_m[name] = ds

    def add_dependency(self, dep):
        self.dependencies.append(dep)

    @staticmethod
    def mkFromProj(proj_dir : str) -> 'ProjInfo':
        ret : ProjInfo = None
        
        # First, see if this is a new-style project
        if os.path.isfile(os.path.join(proj_dir, "ivpm.yaml")):
            note("Reading ivpm.yaml from project %s" % proj_dir)
            path = os.path.join(proj_dir, "ivpm.yaml");
            with open(path, "r") as fp:
                ret = IvpmYamlReader().read(fp, path)
        else:
            # This doesn't appear to be an IVPM project
            # No IVPM-specific data to rely on here
            pass
        return ret

#    @property        
#    def deps(self):
#        return self.dependencies


#****************************************************************************
#* cmd_clone.py
#*
#* Copyright 2025 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#****************************************************************************
import os
import subprocess

from ..utils import fatal
from ..project_ops import ProjectOps


class CmdClone(object):

    def __call__(self, args):
        # Determine workspace directory
        src = args.src
        wsdir = args.workspace_dir

        if wsdir is None:
            # Derive from basename of src (strip trailing .git if present)
            base = os.path.basename(src)
            if base.endswith('.git'):
                base = base[:-4]
            wsdir = base
        
        if os.path.isabs(wsdir):
            target_dir = wsdir
        else:
            target_dir = os.path.abspath(wsdir)
        
        if os.path.exists(target_dir):
            # Allow existing empty directory
            if os.listdir(target_dir):
                fatal("Workspace directory '%s' already exists and is not empty" % target_dir)
        
        # Decide if we are handling a Git source. For now, support Git only.
        self._clone_git(src, target_dir, args)

        # After cloning, run ivpm update in the new workspace
        # so dependencies are fetched according to options provided
        dep_set = getattr(args, 'dep_set', None)
        ivpm_yaml_path = os.path.join(target_dir, "ivpm.yaml")
        
        if os.path.isfile(ivpm_yaml_path):
            ProjectOps(target_dir).update(
                dep_set=dep_set,
                args=args
            )
        else:
            if dep_set is not None:
                fatal("Dependency set '%s' specified but no ivpm.yaml exists in cloned project" % dep_set)
            # No ivpm.yaml and no dep_set specified - just skip update

    def _clone_git(self, src, target_dir, args):
        # Construct clone URL. Convert to SSH form unless anonymous requested
        url = src
        use_anonymous = getattr(args, 'anonymous', False)

        if not use_anonymous:
            # Convert https/http URLs to git@host:path form; keep file: and local paths
            if '://' in src:
                proto = src.split('://', 1)[0]
                if proto != 'file':
                    rest = src.split('://', 1)[1]
                    # Transform host/path to git@host:path
                    first_sl = rest.find('/')
                    if first_sl != -1:
                        host = rest[:first_sl]
                        path = rest[first_sl+1:]
                        url = f"git@{host}:{path}"
            elif src.startswith('git@'):
                url = src
            else:
                # local path - leave as-is
                url = src
        
        git_cmd = ["git", "clone", url, target_dir]
        status = subprocess.run(git_cmd)
        if status.returncode != 0:
            fatal("Git command \"%s\" failed" % str(git_cmd))

        # Handle branch selection/creation
        branch = getattr(args, 'branch', None)
        if branch is not None:
            # Determine if remote branch exists
            # First, fetch to ensure remotes are up to date
            subprocess.run(["git", "fetch", "--all"], cwd=target_dir)
            # Check if origin/branch exists
            have_remote = False
            try:
                out = subprocess.check_output(["git", "ls-remote", "--heads", "origin", branch], cwd=target_dir)
                have_remote = (len(out.decode().strip()) > 0)
            except Exception:
                have_remote = False
            if have_remote:
                # Checkout branch tracking origin/branch
                status = subprocess.run(["git", "checkout", "-B", branch, f"origin/{branch}"], cwd=target_dir)
            else:
                # Create a new local branch from current HEAD
                status = subprocess.run(["git", "checkout", "-b", branch], cwd=target_dir)
            if status.returncode != 0:
                fatal("Failed to checkout branch %s" % branch)



#****************************************************************************
#* package_git.py
#*
#* Copyright 2023 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on:
#*     Author: 
#*
#****************************************************************************
import os
import sys
import subprocess
import dataclasses as dc
from .package_url import PackageURL
from ..proj_info import ProjInfo
from ..project_ops_info import ProjectUpdateInfo, ProjectStatusInfo, ProjectSyncInfo
from ..utils import note, fatal
from ..cache import Cache, is_github_url, parse_github_url

@dc.dataclass
class PackageGit(PackageURL):
    branch : str = None
    commit : str = None
    tag : str = None
    depth : str = None
    anonymous : bool = None

    def update(self, update_info : ProjectUpdateInfo) -> ProjInfo:
        pkg_dir = os.path.join(update_info.deps_dir, self.name)
        self.path = pkg_dir.replace("\\", "/")

        # Report this package for cache statistics
        # Git packages are cacheable if cache=True, editable if cache is not True
        is_cacheable = self.cache is True
        is_editable = self.cache is not True  # Could be cached but isn't
        update_info.report_package(cacheable=is_cacheable, editable=is_editable)

        if os.path.exists(pkg_dir) or os.path.islink(pkg_dir):
            note("package %s is already loaded" % self.name)
        else:
            # Check if caching is enabled and supported
            if self.cache is True:
                # For GitHub URLs, use GitHub API; for others, use git ls-remote
                return self._update_with_cache(update_info, pkg_dir)
            elif self.cache is False:
                # Explicitly no cache - clone without history and make read-only
                return self._update_no_cache_readonly(update_info, pkg_dir)
            else:
                # cache not specified - clone with full history
                return self._update_full_clone(update_info, pkg_dir)

        return ProjInfo.mkFromProj(pkg_dir)

    def _get_github_commit_hash(self, owner: str, repo: str, ref: str = None) -> str:
        """Get the commit hash for a GitHub repo using the API or git ls-remote.
        
        For general git URLs, uses git ls-remote to get the hash.
        """
        import httpx
        
        if ref is None:
            ref = self.branch or self.tag or "HEAD"
        
        # Try GitHub API first if it's a GitHub URL
        if is_github_url(self.url):
            try:
                # Use GitHub API to get the commit hash
                api_url = f"https://api.github.com/repos/{owner}/{repo}/commits/{ref}"
                response = httpx.get(api_url, follow_redirects=True, timeout=30)
                if response.status_code == 200:
                    data = response.json()
                    return data["sha"]
            except Exception:
                pass
        
        # Fallback to git ls-remote for any git URL
        return self._get_commit_hash_ls_remote(ref)
    
    def _get_commit_hash_ls_remote(self, ref: str = None) -> str:
        """Get commit hash using git ls-remote."""
        if ref is None:
            ref = self.branch or self.tag or "HEAD"
        
        try:
            # Use git ls-remote to get the hash
            result = subprocess.run(
                ["git", "ls-remote", self.url, ref],
                capture_output=True,
                text=True,
                timeout=30
            )
            if result.returncode == 0 and result.stdout.strip():
                # Output format: "hash\tref"
                return result.stdout.strip().split()[0]
            
            # Try refs/heads/ prefix for branches
            if not ref.startswith("refs/"):
                result = subprocess.run(
                    ["git", "ls-remote", self.url, f"refs/heads/{ref}"],
                    capture_output=True,
                    text=True,
                    timeout=30
                )
                if result.returncode == 0 and result.stdout.strip():
                    return result.stdout.strip().split()[0]
                
                # Try refs/tags/ prefix for tags
                result = subprocess.run(
                    ["git", "ls-remote", self.url, f"refs/tags/{ref}"],
                    capture_output=True,
                    text=True,
                    timeout=30
                )
                if result.returncode == 0 and result.stdout.strip():
                    return result.stdout.strip().split()[0]
        except Exception:
            pass
        
        return None

    def _update_with_cache(self, update_info: ProjectUpdateInfo, pkg_dir: str) -> ProjInfo:
        """Update using the cache."""
        note("loading package %s with cache" % self.name)
        
        ref = self.branch or self.tag or "HEAD"
        
        # Get the commit hash - use GitHub API for GitHub URLs, git ls-remote otherwise
        commit_hash = None
        if is_github_url(self.url):
            owner, repo = parse_github_url(self.url)
            commit_hash = self._get_github_commit_hash(owner, repo, ref)
        else:
            # Use git ls-remote for general git URLs
            commit_hash = self._get_commit_hash_ls_remote(ref)
        
        if commit_hash is None:
            fatal("Failed to get commit hash for %s (ref: %s)" % (self.url, ref))
        
        cache = update_info.cache
        if cache is None:
            cache = Cache()
        
        # If cache is not properly configured, fall back to full clone
        if not cache.is_enabled():
            note("IVPM_CACHE not set - falling back to full clone for %s" % self.name)
            return self._update_full_clone(update_info, pkg_dir)
        
        # Check if this version is cached
        if cache.has_version(self.name, commit_hash):
            # Cache hit - symlink to deps
            note("Cache hit for %s at %s" % (self.name, commit_hash[:12]))
            cache.link_to_deps(self.name, commit_hash, update_info.deps_dir)
            update_info.report_cache_hit()
            return ProjInfo.mkFromProj(pkg_dir)
        
        # Cache miss - clone without history
        note("Cache miss for %s - cloning" % self.name)
        update_info.report_cache_miss()
        
        # Clone to a temporary location first
        temp_dir = os.path.join(update_info.deps_dir, f".cache_temp_{self.name}")
        if os.path.exists(temp_dir):
            import shutil
            shutil.rmtree(temp_dir)
        
        self._clone_to_dir(update_info, temp_dir, depth=1)
        
        # Store in cache and link
        cache.store_version(self.name, commit_hash, temp_dir)
        cache.link_to_deps(self.name, commit_hash, update_info.deps_dir)
        
        return ProjInfo.mkFromProj(pkg_dir)

    def _update_no_cache_readonly(self, update_info: ProjectUpdateInfo, pkg_dir: str) -> ProjInfo:
        """Clone without history and make read-only (cache=False)."""
        note("loading package %s (no cache, read-only)" % self.name)
        
        self._clone_to_dir(update_info, pkg_dir, depth=1)
        
        # Make read-only
        self._make_readonly(pkg_dir)
        
        return ProjInfo.mkFromProj(pkg_dir)

    def _update_full_clone(self, update_info: ProjectUpdateInfo, pkg_dir: str) -> ProjInfo:
        """Full clone with history (cache unspecified)."""
        note("loading package %s" % self.name)
        
        self._clone_to_dir(update_info, pkg_dir, depth=self.depth)
        
        return ProjInfo.mkFromProj(pkg_dir)

    def _clone_to_dir(self, update_info: ProjectUpdateInfo, target_dir: str, depth=None):
        """Clone the repo to the specified directory."""
        cwd = os.getcwd()
        parent_dir = os.path.dirname(target_dir)
        target_name = os.path.basename(target_dir)
        
        if not os.path.isdir(parent_dir):
            os.makedirs(parent_dir)
        
        os.chdir(parent_dir)
        sys.stdout.flush()

        git_cmd = ["git", "clone"]
    
        if depth is not None:
            git_cmd.extend(["--depth", str(depth)])

        if self.branch is not None:
            git_cmd.extend(["-b", str(self.branch)])

        # Modify the URL to use SSH/key-based clones
        # unless anonymous cloning was requested
        if update_info.args is not None and hasattr(update_info.args, "anonymous"):
            use_anonymous = getattr(update_info.args, "anonymous")
        else:
            use_anonymous = False

        if self.anonymous is not None:
            use_anonymous = self.anonymous

        if not use_anonymous:
            print("NOTE: using dev URL")
            delim_idx = self.url.find("://")
            protocol = self.url[:delim_idx]
            if protocol != "file":
                url = self.url[delim_idx+3:]
                first_sl_idx = url.find('/')
                url = "git@" + url[:first_sl_idx] + ":" + url[first_sl_idx+1:]
                print("Final URL: %s" % url)
                git_cmd.append(url)
            else:
                print("NOTE: using original file-based URL")
                git_cmd.append(self.url)
        else:
            print("NOTE: using anonymous URL")
            git_cmd.append(self.url)

        # Clone to the target directory name
        git_cmd.append(target_name)
        
        print("git_cmd: \"" + str(git_cmd) + "\"")
        status = subprocess.run(git_cmd)
        os.chdir(cwd)
    
        if status.returncode != 0:
            fatal("Git command \"%s\" failed" % str(git_cmd))

        # Checkout a specific commit            
        if self.commit is not None:
            os.chdir(target_dir)
            git_cmd = "git reset --hard %s" % self.commit
            status = os.system(git_cmd)
        
            if status != 0:
                fatal("Git command \"%s\" failed" % str(git_cmd))
            os.chdir(cwd)
        
    
        # TODO: Existence of .gitmodules should trigger this
        if os.path.isfile(os.path.join(target_dir, ".gitmodules")):
            os.chdir(target_dir)
            sys.stdout.flush()
            status = os.system("git submodule update --init --recursive")
            os.chdir(cwd)

    def _make_readonly(self, path: str):
        """Make all files in a directory tree read-only."""
        import stat
        for root, dirs, files in os.walk(path):
            for d in dirs:
                dir_path = os.path.join(root, d)
                mode = os.stat(dir_path).st_mode
                os.chmod(dir_path, mode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)
            for f in files:
                file_path = os.path.join(root, f)
                mode = os.stat(file_path).st_mode
                os.chmod(file_path, mode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)
        mode = os.stat(path).st_mode
        os.chmod(path, mode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)
    
    def status(self, status_info : ProjectStatusInfo):
        pass
    
    def sync(self, sync_info : ProjectSyncInfo):
        if not os.path.isdir(os.path.join(sync_info.deps_dir, dir, ".git")):
            fatal("Package \"" + dir + "\" is not a Git repository")
        print("Package: " + dir)
        try:
            branch = subprocess.check_output(
                ["git", "branch"],
                cwd=os.path.join(sync_info.deps_dir, dir))
        except Exception as e:
            fatal("failed to get branch of package \"" + dir + "\"")

        branch = branch.strip()
        if len(branch) == 0:
            fatal("branch is empty")

        branch_lines = branch.decode().splitlines()
        branch = None
        for bl in branch_lines:
            if bl[0] == "*":
                branch = bl[1:].strip()
                break
        if branch is None:
            fatal("Failed to identify branch")

        status = subprocess.run(
            ["git", "fetch"],
            cwd=os.path.join(sync_info.deps_dir, dir))
        if status.returncode != 0:
            fatal("Failed to run git fetch on package %s" % dir)
        status = subprocess.run(
            ["git", "merge", "origin/" + branch],
            cwd=os.path.join(sync_info.deps_dir, dir))
        if status.returncode != 0:
            fatal("Failed to run git merge origin/%s on package %s" % (branch, dir))
    
    def process_options(self, opts, si):
        super().process_options(opts, si)

        if "anonymous" in opts.keys():
            self.anonymous = opts["anonymous"]
                
        if "depth" in opts.keys():
            self.depth = opts["depth"]
                
        if "dep-set" in opts.keys():
            self.dep_set = opts["dep-set"]
               
        if "branch" in opts.keys():
            self.branch = opts["branch"]
                
        if "commit" in opts.keys():
            self.commit = opts["commit"]
               
        if "tag" in opts.keys():
            self.tag = opts["tag"]


    @staticmethod
    def create(name, opts, si) -> 'PackageGit':
        pkg = PackageGit(name)
        pkg.process_options(opts, si)
        return pkg




import os
import shutil
import stat
import subprocess
import tempfile
import unittest
from unittest.mock import patch, MagicMock

from .test_base import TestBase

ROOTDIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
SRCDIR = os.path.join(ROOTDIR, 'src')

import sys
sys.path.insert(0, SRCDIR)

from ivpm.cache import Cache, is_github_url, parse_github_url, CacheResult
from ivpm.project_ops_info import ProjectUpdateInfo


class TestCacheHelpers(unittest.TestCase):
    """Test helper functions in cache module."""
    
    def test_is_github_url_https(self):
        self.assertTrue(is_github_url("https://github.com/owner/repo.git"))
        self.assertTrue(is_github_url("https://github.com/owner/repo"))
        
    def test_is_github_url_git(self):
        self.assertTrue(is_github_url("git@github.com:owner/repo.git"))
        
    def test_is_github_url_false(self):
        self.assertFalse(is_github_url("https://gitlab.com/owner/repo.git"))
        self.assertFalse(is_github_url("https://bitbucket.org/owner/repo.git"))
        
    def test_parse_github_url_https(self):
        owner, repo = parse_github_url("https://github.com/fvutils/vlsim.git")
        self.assertEqual(owner, "fvutils")
        self.assertEqual(repo, "vlsim")
        
    def test_parse_github_url_https_no_git(self):
        owner, repo = parse_github_url("https://github.com/fvutils/vlsim")
        self.assertEqual(owner, "fvutils")
        self.assertEqual(repo, "vlsim")
        
    def test_parse_github_url_ssh(self):
        owner, repo = parse_github_url("git@github.com:fvutils/vlsim.git")
        self.assertEqual(owner, "fvutils")
        self.assertEqual(repo, "vlsim")
        
    def test_parse_github_url_with_path(self):
        owner, repo = parse_github_url("https://github.com/fvutils/vlsim/tree/main")
        self.assertEqual(owner, "fvutils")
        self.assertEqual(repo, "vlsim")
        
    def test_parse_github_url_invalid(self):
        owner, repo = parse_github_url("https://gitlab.com/owner/repo.git")
        self.assertIsNone(owner)
        self.assertIsNone(repo)


class TestCache(unittest.TestCase):
    """Test Cache class."""
    
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.cache_dir = os.path.join(self.test_dir, "cache")
        self.deps_dir = os.path.join(self.test_dir, "deps")
        os.makedirs(self.deps_dir)
        self.cache = Cache(self.cache_dir)
        
    def tearDown(self):
        # Need to make files writable before removal
        for root, dirs, files in os.walk(self.test_dir):
            for d in dirs:
                dir_path = os.path.join(root, d)
                try:
                    os.chmod(dir_path, stat.S_IRWXU)
                except:
                    pass
            for f in files:
                file_path = os.path.join(root, f)
                try:
                    os.chmod(file_path, stat.S_IRWXU)
                except:
                    pass
        shutil.rmtree(self.test_dir)
        
    def test_get_package_cache_dir(self):
        path = self.cache.get_package_cache_dir("mypackage")
        self.assertEqual(path, os.path.join(self.cache_dir, "mypackage"))
        
    def test_get_version_cache_dir(self):
        path = self.cache.get_version_cache_dir("mypackage", "abc123")
        self.assertEqual(path, os.path.join(self.cache_dir, "mypackage", "abc123"))
        
    def test_has_version_false(self):
        self.assertFalse(self.cache.has_version("mypackage", "abc123"))
        
    def test_has_version_true(self):
        version_dir = self.cache.get_version_cache_dir("mypackage", "abc123")
        os.makedirs(version_dir)
        self.assertTrue(self.cache.has_version("mypackage", "abc123"))
        
    def test_ensure_cache_dir(self):
        pkg_dir = self.cache.ensure_cache_dir("mypackage")
        self.assertTrue(os.path.isdir(pkg_dir))
        self.assertEqual(pkg_dir, self.cache.get_package_cache_dir("mypackage"))
        
    def test_store_version(self):
        # Create a source directory
        source_dir = os.path.join(self.test_dir, "source")
        os.makedirs(source_dir)
        with open(os.path.join(source_dir, "test.txt"), "w") as f:
            f.write("test content")
            
        # Store in cache
        cached_path = self.cache.store_version("mypackage", "abc123", source_dir)
        
        # Verify
        self.assertTrue(os.path.isdir(cached_path))
        self.assertTrue(os.path.isfile(os.path.join(cached_path, "test.txt")))
        self.assertFalse(os.path.exists(source_dir))  # Source was moved
        
        # Verify read-only
        test_file = os.path.join(cached_path, "test.txt")
        mode = os.stat(test_file).st_mode
        self.assertFalse(mode & stat.S_IWUSR)
        
    def test_link_to_deps(self):
        # Create a cached version
        version_dir = self.cache.get_version_cache_dir("mypackage", "abc123")
        os.makedirs(version_dir)
        with open(os.path.join(version_dir, "test.txt"), "w") as f:
            f.write("test content")
            
        # Link to deps
        link_path = self.cache.link_to_deps("mypackage", "abc123", self.deps_dir)
        
        # Verify
        self.assertTrue(os.path.islink(link_path))
        self.assertEqual(os.path.realpath(link_path), version_dir)
        self.assertTrue(os.path.isfile(os.path.join(link_path, "test.txt")))
        
    def test_link_to_deps_replaces_existing(self):
        # Create existing directory
        existing = os.path.join(self.deps_dir, "mypackage")
        os.makedirs(existing)
        with open(os.path.join(existing, "old.txt"), "w") as f:
            f.write("old content")
            
        # Create a cached version
        version_dir = self.cache.get_version_cache_dir("mypackage", "abc123")
        os.makedirs(version_dir)
        with open(os.path.join(version_dir, "new.txt"), "w") as f:
            f.write("new content")
            
        # Link to deps
        link_path = self.cache.link_to_deps("mypackage", "abc123", self.deps_dir)
        
        # Verify old directory replaced with link
        self.assertTrue(os.path.islink(link_path))
        self.assertTrue(os.path.isfile(os.path.join(link_path, "new.txt")))
        self.assertFalse(os.path.exists(os.path.join(link_path, "old.txt")))
        
    def test_get_cache_info_empty(self):
        info = self.cache.get_cache_info()
        self.assertEqual(info["packages"], [])
        self.assertEqual(info["total_size"], 0)
        
    def test_get_cache_info_with_data(self):
        # Create some cached packages
        v1 = self.cache.get_version_cache_dir("pkg1", "v1")
        os.makedirs(v1)
        with open(os.path.join(v1, "test.txt"), "w") as f:
            f.write("content")
            
        v2 = self.cache.get_version_cache_dir("pkg1", "v2")
        os.makedirs(v2)
        with open(os.path.join(v2, "test.txt"), "w") as f:
            f.write("more content")
            
        info = self.cache.get_cache_info()
        self.assertEqual(len(info["packages"]), 1)
        self.assertEqual(info["packages"][0]["name"], "pkg1")
        self.assertEqual(len(info["packages"][0]["versions"]), 2)
        self.assertGreater(info["total_size"], 0)
        
    def test_clean_older_than(self):
        import time
        
        # Create a cached version
        v1 = self.cache.get_version_cache_dir("pkg1", "v1")
        os.makedirs(v1)
        with open(os.path.join(v1, "test.txt"), "w") as f:
            f.write("content")
        self.cache._make_readonly(v1)
        
        # Set modification time to 10 days ago
        old_time = time.time() - (10 * 24 * 60 * 60)
        os.utime(v1, (old_time, old_time))
        
        # Create a recent version
        v2 = self.cache.get_version_cache_dir("pkg1", "v2")
        os.makedirs(v2)
        with open(os.path.join(v2, "test.txt"), "w") as f:
            f.write("new content")
            
        # Clean entries older than 7 days
        removed = self.cache.clean_older_than(7)
        
        self.assertEqual(removed, 1)
        self.assertFalse(self.cache.has_version("pkg1", "v1"))
        self.assertTrue(self.cache.has_version("pkg1", "v2"))


class TestCacheGit(TestBase):
    """Test git caching integration."""
    
    def setUp(self):
        super().setUp()
        # Set up IVPM_CACHE for caching tests
        self.cache_dir = os.path.join(self.testdir, ".ivpm_cache")
        os.makedirs(self.cache_dir, exist_ok=True)
        os.environ["IVPM_CACHE"] = self.cache_dir
    
    def tearDown(self):
        # Clean up IVPM_CACHE env var
        if "IVPM_CACHE" in os.environ:
            del os.environ["IVPM_CACHE"]
        super().tearDown()
    
    def _init_git_repo(self, path, files=None, branch='main'):
        """Initialize a git repo for testing."""
        os.makedirs(path, exist_ok=True)
        subprocess.check_call(["git", "init", "-b", branch], cwd=path)
        subprocess.check_call(["git", "config", "user.email", "test@example.com"], cwd=path)
        subprocess.check_call(["git", "config", "user.name", "Test"], cwd=path)
        if files:
            for rel, content in files.items():
                full = os.path.join(path, rel)
                os.makedirs(os.path.dirname(full), exist_ok=True) if os.path.dirname(full) else None
                with open(full, 'w') as f:
                    f.write(content)
        if not os.path.exists(os.path.join(path, 'ivpm.yaml')):
            with open(os.path.join(path, 'ivpm.yaml'), 'w') as f:
                f.write('package:\n  name: sample\n  dep-sets:\n    - name: default-dev\n      deps: []\n')
        subprocess.check_call(["git", "add", "-A"], cwd=path)
        subprocess.check_call(["git", "commit", "-m", "init"], cwd=path)
    
    def test_git_cache_local_file_url(self):
        """Test that cache=true works for local file:// git URLs using git ls-remote."""
        src_repo = os.path.join(self.testdir, 'src_repo')
        self._init_git_repo(src_repo, files={"test.txt": "test content"})
        
        self.mkFile("ivpm.yaml", f"""
        package:
            name: cache_test
            dep-sets:
                - name: default-dev
                  deps:
                    - name: test_pkg
                      url: file://{src_repo}
                      src: git
                      cache: true
        """)
        
        self.ivpm_update(skip_venv=True)
        
        pkg_dir = os.path.join(self.testdir, "packages", "test_pkg")
        # Should be a symlink to cache
        self.assertTrue(os.path.islink(pkg_dir))
        self.assertTrue(os.path.isfile(os.path.join(pkg_dir, "test.txt")))
        
        # Verify files are read-only since cached
        test_file = os.path.join(pkg_dir, "test.txt")
        mode = os.stat(test_file).st_mode
        self.assertFalse(mode & stat.S_IWUSR)
    
    def test_git_no_cache_readonly(self):
        """Test that cache=false makes files read-only without cache."""
        src_repo = os.path.join(self.testdir, 'src_repo')
        self._init_git_repo(src_repo, files={"test.txt": "test content"})
        
        self.mkFile("ivpm.yaml", f"""
        package:
            name: cache_test
            dep-sets:
                - name: default-dev
                  deps:
                    - name: test_pkg
                      url: file://{src_repo}
                      src: git
                      cache: false
        """)
        
        self.ivpm_update(skip_venv=True)
        
        pkg_dir = os.path.join(self.testdir, "packages", "test_pkg")
        self.assertTrue(os.path.isdir(pkg_dir))
        
        # Verify files are read-only
        test_file = os.path.join(pkg_dir, "test.txt")
        mode = os.stat(test_file).st_mode
        self.assertFalse(mode & stat.S_IWUSR)
        
    def test_git_unspecified_cache_full_clone(self):
        """Test that cache unspecified does full clone with write access."""
        src_repo = os.path.join(self.testdir, 'src_repo')
        self._init_git_repo(src_repo, files={"test.txt": "test content"})
        
        self.mkFile("ivpm.yaml", f"""
        package:
            name: cache_test
            dep-sets:
                - name: default-dev
                  deps:
                    - name: test_pkg
                      url: file://{src_repo}
                      src: git
        """)
        
        self.ivpm_update(skip_venv=True)
        
        pkg_dir = os.path.join(self.testdir, "packages", "test_pkg")
        self.assertTrue(os.path.isdir(pkg_dir))
        self.assertTrue(os.path.isdir(os.path.join(pkg_dir, ".git")))
        
        # Verify files are writable (default behavior)
        test_file = os.path.join(pkg_dir, "test.txt")
        mode = os.stat(test_file).st_mode
        self.assertTrue(mode & stat.S_IWUSR)


class TestCacheHttp(TestBase):
    """Test HTTP URL caching."""
    
    def setUp(self):
        super().setUp()
        # Set up IVPM_CACHE for caching tests
        self.cache_dir = os.path.join(self.testdir, ".ivpm_cache")
        os.makedirs(self.cache_dir, exist_ok=True)
        os.environ["IVPM_CACHE"] = self.cache_dir
    
    def tearDown(self):
        # Clean up IVPM_CACHE env var
        if "IVPM_CACHE" in os.environ:
            del os.environ["IVPM_CACHE"]
        super().tearDown()
    
    def test_http_cache_true(self):
        """Test that cache=true works for HTTP URLs."""
        self.mkFile("ivpm.yaml", """
        package:
            name: http_cache_test
            dep-sets:
                - name: default-dev
                  deps:
                    - name: googletest
                      url: https://github.com/google/googletest/archive/refs/tags/v1.15.2.tar.gz
                      cache: true
        """)
        
        self.ivpm_update(skip_venv=True)
        
        pkg_dir = os.path.join(self.testdir, "packages", "googletest")
        # Should be a symlink to cache
        self.assertTrue(os.path.islink(pkg_dir) or os.path.isdir(pkg_dir))
        self.assertTrue(os.path.isfile(os.path.join(pkg_dir, "README.md")))
        
    def test_http_no_cache_readonly(self):
        """Test that cache=false makes HTTP downloads read-only."""
        self.mkFile("ivpm.yaml", """
        package:
            name: http_cache_test
            dep-sets:
                - name: default-dev
                  deps:
                    - name: googletest
                      url: https://github.com/google/googletest/archive/refs/tags/v1.15.2.tar.gz
                      cache: false
        """)
        
        self.ivpm_update(skip_venv=True)
        
        pkg_dir = os.path.join(self.testdir, "packages", "googletest")
        self.assertTrue(os.path.isdir(pkg_dir))
        
        # Verify files are read-only
        readme = os.path.join(pkg_dir, "README.md")
        mode = os.stat(readme).st_mode
        self.assertFalse(mode & stat.S_IWUSR)


class TestCacheGitHub(TestBase):
    """Test GitHub git URL caching (requires network)."""
    
    def setUp(self):
        super().setUp()
        # Set up IVPM_CACHE for caching tests
        self.cache_dir = os.path.join(self.testdir, ".ivpm_cache")
        os.makedirs(self.cache_dir, exist_ok=True)
        os.environ["IVPM_CACHE"] = self.cache_dir
    
    def tearDown(self):
        # Clean up IVPM_CACHE env var
        if "IVPM_CACHE" in os.environ:
            del os.environ["IVPM_CACHE"]
        super().tearDown()
    
    def test_github_cache_creates_symlink(self):
        """Test that cache=true for GitHub URLs creates a symlink."""
        self.mkFile("ivpm.yaml", """
        package:
            name: github_cache_test
            dep-sets:
                - name: default-dev
                  deps:
                    - name: vlsim
                      url: https://github.com/fvutils/vlsim.git
                      anonymous: true
                      cache: true
        """)
        
        self.ivpm_update(skip_venv=True)
        
        pkg_dir = os.path.join(self.testdir, "packages", "vlsim")
        # Should be a symlink to cache
        self.assertTrue(os.path.islink(pkg_dir))
        self.assertTrue(os.path.isfile(os.path.join(pkg_dir, "README.md")))
        
        # Verify files are read-only since cached
        readme = os.path.join(pkg_dir, "README.md")
        mode = os.stat(readme).st_mode
        self.assertFalse(mode & stat.S_IWUSR)
    
    def test_github_cache_hit(self):
        """Test that cache hit scenario works (second update uses cached version)."""
        # First update - creates cache entry
        self.mkFile("ivpm.yaml", """
        package:
            name: github_cache_test
            dep-sets:
                - name: default-dev
                  deps:
                    - name: vlsim
                      url: https://github.com/fvutils/vlsim.git
                      anonymous: true
                      branch: master
                      cache: true
        """)
        
        self.ivpm_update(skip_venv=True)
        
        # Verify cached
        pkg_dir = os.path.join(self.testdir, "packages", "vlsim")
        self.assertTrue(os.path.islink(pkg_dir))
        
        # Record the cache target
        cache_target = os.path.realpath(pkg_dir)
        
        # Remove the symlink but keep the cache
        os.unlink(pkg_dir)
        
        # Second update - should use cached version
        self.ivpm_update(skip_venv=True)
        
        # Should be same cache target
        self.assertTrue(os.path.islink(pkg_dir))
        self.assertEqual(os.path.realpath(pkg_dir), cache_target)


class TestCacheGeneralGit(TestBase):
    """Test general git URL support (non-GitHub) using git ls-remote."""
    
    def _init_git_repo(self, path, files=None, branch='main'):
        """Initialize a git repo for testing."""
        os.makedirs(path, exist_ok=True)
        subprocess.check_call(["git", "init", "-b", branch], cwd=path)
        subprocess.check_call(["git", "config", "user.email", "test@example.com"], cwd=path)
        subprocess.check_call(["git", "config", "user.name", "Test"], cwd=path)
        if files:
            for rel, content in files.items():
                full = os.path.join(path, rel)
                dirname = os.path.dirname(full)
                if dirname and not os.path.isdir(dirname):
                    os.makedirs(dirname)
                with open(full, 'w') as f:
                    f.write(content)
        if not os.path.exists(os.path.join(path, 'ivpm.yaml')):
            with open(os.path.join(path, 'ivpm.yaml'), 'w') as f:
                f.write('package:\n  name: sample\n  dep-sets:\n    - name: default-dev\n      deps: []\n')
        subprocess.check_call(["git", "add", "-A"], cwd=path)
        subprocess.check_call(["git", "commit", "-m", "init"], cwd=path)
    
    def test_git_ls_remote_hash_retrieval(self):
        """Test that git ls-remote can be used to get hash for local repos."""
        from ivpm.pkg_types.package_git import PackageGit
        
        src_repo = os.path.join(self.testdir, 'src_repo')
        self._init_git_repo(src_repo, files={"test.txt": "content"})
        
        # Get expected hash
        result = subprocess.run(
            ["git", "rev-parse", "HEAD"],
            cwd=src_repo,
            capture_output=True,
            text=True
        )
        expected_hash = result.stdout.strip()
        
        # Test PackageGit can get hash
        pkg = PackageGit(name="test", url=f"file://{src_repo}")
        retrieved_hash = pkg._get_commit_hash_ls_remote("HEAD")
        
        self.assertEqual(retrieved_hash, expected_hash)


class TestProjectUpdateInfoCache(unittest.TestCase):
    """Test ProjectUpdateInfo cache hit/miss tracking."""
    
    def test_cache_hit_tracking(self):
        update_info = ProjectUpdateInfo(None, "/tmp/deps")
        self.assertEqual(update_info.cache_hits, 0)
        self.assertEqual(update_info.cache_misses, 0)
        
        update_info.report_cache_hit()
        self.assertEqual(update_info.cache_hits, 1)
        
        update_info.report_cache_hit()
        self.assertEqual(update_info.cache_hits, 2)
        
    def test_cache_miss_tracking(self):
        update_info = ProjectUpdateInfo(None, "/tmp/deps")
        
        update_info.report_cache_miss()
        self.assertEqual(update_info.cache_misses, 1)


if __name__ == '__main__':
    unittest.main()



#****************************************************************************
#* pkg_compile_flags.py
#*
#* Copyright 2022 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on:
#*     Author: 
#*
#****************************************************************************
from .pkg_info import PkgInfo

class PkgCompileFlags(object):

    def __init__(self):
        self._processed_pkgs = set()
        pass

    def cflags(self, pkg : PkgInfo):
        ret = []
        self._processed_pkgs.clear()
        for p in pkg:
            self._collect_cflags(ret, p)

        return ret

    def lflags(self, pkg : PkgInfo):
        ret = []
        self._processed_pkgs.clear()
        for p in pkg:
            self._collect_lflags(ret, p)

        return ret

    def flags(self, pkg):
        ret = []
        self._processed_pkgs.clear()
        for p in pkg:
            self._collect_cflags(ret, p)

        self._processed_pkgs.clear()
        for p in pkg:
            self._collect_lflags(ret, p)

        return ret

    def libs(self, pkg, kind):
        ret = []
        self._processed_pkgs.clear()
        for p in pkg:
            self._collect_libs(ret, p, kind)

        return ret

    def paths(self, pkg, kind):
        ret = []
        self._processed_pkgs.clear()
        for p in pkg:
            self._collect_paths(ret, p, kind)

        return ret
    
    def ldirs(self, pkg : PkgInfo, kind):
        ret = []
        self._processed_pkgs.clear()
        self._collect_ldirs(ret, pkg, kind)
        return ret

    def _collect_cflags(self, ret, pkg_l : PkgInfo):
        for pkg in pkg_l:
            for i in pkg._incdirs:
                inc = "-I" + i
                if inc not in ret:
                    ret.append(inc)

        for d in pkg._deps:
            if d._name not in self._processed_pkgs:
                self._processed_pkgs.add(d._name)
                self._collect_cflags(ret, d)

    def _collect_paths(self, ret, pkg_l : PkgInfo, kind):
        for pkg in pkg_l:
            for path in pkg.getPaths(kind):
                if path not in ret:
                    ret.append(path)

            for d in pkg._deps:
                if d._name not in self._processed_pkgs:
                    self._processed_pkgs.add(d._name)
                    self._collect_paths(ret, d, kind)

    def _collect_libs(self, ret, pkg_l : PkgInfo, kind):
        for pkg in pkg_l:
            libs = pkg.getLibs(kind)
            for path in [] if libs is None else libs:
                if path not in ret:
                    ret.append(path)

            for d in pkg._deps:
                if d._name not in self._processed_pkgs:
                    self._processed_pkgs.add(d._name)
                    self._collect_libs(ret, d, kind)

    def _collect_ldirs(self, ret, pkg_l : PkgInfo, kind):
        for pkg in pkg_l:
            for d in pkg.getLibDirs(kind):
                if d not in ret:
                    ret.append(d)

            for d in pkg._deps:
                if d._name not in self._processed_pkgs:
                    self._processed_pkgs.add(d._name)
                    self._collect_ldirs(ret, d, kind)

    def _collect_lflags(self, ret, pkg : PkgInfo):
        for d in pkg._libdirs:
            lpath = "-L" + d
            if lpath not in ret:
                ret.append(lpath)

        for l in pkg._libs:
            lpath = "-l" + l
            if lpath not in ret:
                ret.append(lpath)

        for d in pkg._deps:
            if d._name not in self._processed_pkgs:
                self._processed_pkgs.add(d._name)
                self._collect_lflags(ret, d)






#****************************************************************************
#* pkg_info_rgy.py
#*
#* Copyright 2023 Matthew Ballance and Contributors
#*
#* Licensed under the Apache License, Version 2.0 (the "License"); you may 
#* not use this file except in compliance with the License.  
#* You may obtain a copy of the License at:
#*
#*   http://www.apache.org/licenses/LICENSE-2.0
#*
#* Unless required by applicable law or agreed to in writing, software 
#* distributed under the License is distributed on an "AS IS" BASIS, 
#* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  
#* See the License for the specific language governing permissions and 
#* limitations under the License.
#*
#* Created on:
#*     Author: 
#*
#****************************************************************************
import os
import sys
import importlib
from typing import Callable

class PkgInfoRgy(object):

    _inst = None

    def __init__(self):
        self.info_m = {}
        pass

    def load(self):
        if sys.version_info < (3,10):
            from importlib_metadata import entry_points
        else:
            from importlib.metadata import entry_points

        plugins = entry_points(group='ivpm.pkginfo')

        for p in plugins:
            try:
                ext_t = p.load()
                ext = ext_t()
                if ext.name not in self.info_m.keys():
                    self.info_m[ext.name] = ext
                else:
                    raise Exception("Duplicate package %s" % ext.name)
            except Exception as e:
                print("IVPM: failed to load plugin (%s)" % str(e))
            
        # Finally, iterate through the path looking for leftovers (?)
        for path in sys.path:
            if os.path.isdir(path):
                for pkg in os.listdir(path):
                    if os.path.isfile(os.path.join(path, pkg, "pkginfo.py")):
                        try:
                            pkginfo_m = importlib.import_module("%s.pkginfo" % pkg)
                        except ModuleNotFoundError as e:
                            sys.stderr.write("Note: package @ %s has pkginfo.py, but cannot load %s.pkginfo (%s)\n" % (path, pkg, str(e)))
                            continue
                        if not hasattr(pkginfo_m, "PkgInfo"):
                            sys.stderr.write("Note: pkginfo (%s) module exists, but is missing PkgInfo\n", os.path.join(path, pkg))
                            continue
                        pkginfo = getattr(pkginfo_m, "PkgInfo")()
                        if pkginfo.name not in self.info_m.keys():
                            self.info_m[pkginfo.name] = pkginfo

    def hasPkg(self, pkg):
        return pkg in self.info_m.keys()
    
    def getPkgs(self):
        return self.info_m.keys()

    def getPkg(self, pkg):
        if pkg in self.info_m.keys():
            return self.info_m[pkg]
        else:
            raise Exception("Package %s is not present" % pkg)

    def getIncDirs(self, kind=None, filter : Callable=None):
        incdirs = []

        for pkg_id in self.info_m.keys():
            pkg = self.info_m[pkg_id]
            if filter is not None and not filter(pkg.name):
                continue
            pkg_incdirs = pkg.getIncDirs()

            if pkg_incdirs is not None:
                for ld in pkg_incdirs:
                    if ld not in incdirs:
                        incdirs.append(ld)

        return incdirs
        
    def getLibDirs(self, kind=None, filter : Callable=None):
        libdirs = []

        for pkg_id in self.info_m.keys():
            pkg = self.info_m[pkg_id]
            if filter is not None and not filter(pkg.name):
                continue
            pkg_libdirs = pkg.getLibDirs(kind)

            if pkg_libdirs is not None:
                for ld in pkg_libdirs:
                    if ld not in libdirs:
                        libdirs.append(ld)

        return libdirs

    def getLibs(self, kind=None, filter : Callable=None):
        libs = []

        for pkg_id in self.info_m.keys():
            pkg = self.info_m[pkg_id]
            if filter is not None and not filter(pkg.name):
                continue
            pkg_libs = pkg.getLibs(kind)

            if pkg_libs is not None:
                for lib in pkg_libs:
                    if lib not in libs:
                        libs.append(lib)

        return libs

    def getPaths(self, kind):
        ret = []
        for name,pkg in self.info_m.items():
            ret.extend(pkg.getPaths(kind))
        return ret

    @classmethod
    def inst(cls):
        if cls._inst is None:
            cls._inst = PkgInfoRgy()
            cls._inst.load()
        return cls._inst





import os
import platform
import sys
import subprocess

ps = ";" if platform.system() == "Windows" else ":"
env = os.environ.copy()
env["IVPM_PYTHONPATH"] = ps.join(sys.path)
exec_dir = os.path.dirname(sys.executable)

if platform.system() == "Windows":
    # User scripts are in 'Scripts'
    env["PATH"] = os.path.join(exec_dir, "Scripts") + ps + env["PATH"]
else:
    # User scripts are alongside the executable
    env["PATH"] = exec_dir + ps + env["PATH"]

def main():
    cmd = sys.argv[1:]

    status = subprocess.run(cmd, env=env)

    return status.returncode

if __name__ == "__main__":
    main()
